Type: Package
Package: CuratedAtlasQueryR
Title: Queries the Human Cell Atlas
Version: 1.10.0
Authors@R: c(
    person(
        "Stefano",
        "Mangiola",
        email = "mangiolastefano@gmail.com",
        role = c("aut", "cre", "rev"),
        comment = c(ORCID = "0000-0001-7474-836X")
    ),
    person(
        "Michael",
        "Milton",
        email = "milton.m@wehi.edu.au",
        role = c("aut", "rev"),
        comment = c(ORCID = "0000-0002-8965-2595")
    ),
    person(
        "Martin",
        "Morgan",
        email = "Martin.Morgan@RoswellPark.org",
        role = c("ctb", "rev")
    ),
    person(
        "Vincent",
        "Carey",
        email = "stvjc@channing.harvard.edu",
        role = c("ctb", "rev")
    ),
    person(
        "Julie",
        "Iskander",
        email = "iskander.j@wehi.edu.au",
        role = c( "rev")
    ),
    person(
        "Tony",
        "Papenfuss",
        email = "papenfuss@wehi.edu.au",
        role = c( "rev")
    ),
    person(
        "Silicon Valley Foundation",
        "CZF2019-002443",
        role = c( "fnd")
    ),
    person(
        "NIH NHGRI",
        "5U24HG004059-18",
        role = c( "fnd")
    ),
    person(
        "Victoria Cancer Agency",
        "ECRF21036",
        role = c( "fnd")
    ),
    person(
        "NHMRC",
        "1116955",
        role = c( "fnd")
    ))
Description: Provides access to a copy of the Human Cell Atlas, but with 
    harmonised metadata. This allows for uniform querying across numerous 
    datasets within the Atlas using common fields such as cell type, tissue 
    type, and patient ethnicity. Usage involves first querying the metadata
    table for cells of interest, and then downloading the corresponding cells 
    into a SingleCellExperiment object. 
License: GPL-3
Depends: R (>= 4.2.0)
Imports: dplyr, SummarizedExperiment, SingleCellExperiment, purrr (>=
        1.0.0), BiocGenerics, glue, HDF5Array, DBI, tools, httr, cli,
        assertthat, SeuratObject, Seurat, methods, rlang, stats,
        S4Vectors, tibble, utils, dbplyr (>= 2.3.0), duckdb, stringr
Suggests: zellkonverter, rmarkdown, knitr, testthat, basilisk, arrow,
        reticulate, spelling, forcats, ggplot2,
        tidySingleCellExperiment, rprojroot
Biarch: true
biocViews: AssayDomain, Infrastructure, RNASeq, DifferentialExpression,
        GeneExpression, Normalization, Clustering, QualityControl,
        Sequencing, Transcription, Transcriptomics
Encoding: UTF-8
RoxygenNote: 7.2.3
URL: https://github.com/stemangiola/CuratedAtlasQueryR
BugReports: https://github.com/stemangiola/CuratedAtlasQueryR/issues
VignetteBuilder: knitr
Roxygen: list(markdown = TRUE)
Collate: 'utils.R' 'counts.R' 'dev.R' 'metadata.R' 'seurat.R'
        'unharmonised.R' 'zzz.R'
Language: en-US
PackageStatus: Deprecated
git_url: https://git.bioconductor.org/packages/CuratedAtlasQueryR
git_branch: RELEASE_3_23
git_last_commit: 553d9af
git_last_commit_date: 2026-04-28
Repository: Bioconductor 3.23
Date/Publication: 2026-04-28
NeedsCompilation: no
Packaged: 2026-04-28 21:31:55 UTC; biocbuild
Author: Stefano Mangiola [aut, cre, rev] (ORCID:
    <https://orcid.org/0000-0001-7474-836X>),
  Michael Milton [aut, rev] (ORCID:
    <https://orcid.org/0000-0002-8965-2595>),
  Martin Morgan [ctb, rev],
  Vincent Carey [ctb, rev],
  Julie Iskander [rev],
  Tony Papenfuss [rev],
  Silicon Valley Foundation CZF2019-002443 [fnd],
  NIH NHGRI 5U24HG004059-18 [fnd],
  Victoria Cancer Agency ECRF21036 [fnd],
  NHMRC 1116955 [fnd]
Maintainer: Stefano Mangiola <mangiolastefano@gmail.com>

# Generated by roxygen2: do not edit by hand

S3method(as.sparse,DelayedMatrix)
export(DATABASE_URL)
export(SAMPLE_DATABASE_URL)
export(get_SingleCellExperiment)
export(get_metadata)
export(get_seurat)
export(get_single_cell_experiment)
export(get_unharmonised_metadata)
importFrom(BiocGenerics,cbind)
importFrom(DBI,dbConnect)
importFrom(DBI,dbDisconnect)
importFrom(HDF5Array,HDF5RealizationSink)
importFrom(HDF5Array,loadHDF5SummarizedExperiment)
importFrom(S4Vectors,DataFrame)
importFrom(Seurat,as.SingleCellExperiment)
importFrom(SeuratObject,as.Seurat)
importFrom(SeuratObject,as.sparse)
importFrom(SingleCellExperiment,SingleCellExperiment)
importFrom(SingleCellExperiment,combineCols)
importFrom(SummarizedExperiment,"assayNames<-")
importFrom(SummarizedExperiment,"colData<-")
importFrom(SummarizedExperiment,colData)
importFrom(assertthat,assert_that)
importFrom(assertthat,has_name)
importFrom(cli,cli_abort)
importFrom(cli,cli_alert_info)
importFrom(cli,cli_alert_success)
importFrom(cli,cli_alert_warning)
importFrom(cli,hash_sha256)
importFrom(dbplyr,remote_con)
importFrom(dbplyr,sql)
importFrom(dplyr,as_tibble)
importFrom(dplyr,collect)
importFrom(dplyr,filter)
importFrom(dplyr,group_by)
importFrom(dplyr,inner_join)
importFrom(dplyr,mutate)
importFrom(dplyr,pull)
importFrom(dplyr,summarise)
importFrom(dplyr,tbl)
importFrom(dplyr,transmute)
importFrom(duckdb,duckdb)
importFrom(glue,glue)
importFrom(httr,GET)
importFrom(httr,HEAD)
importFrom(httr,modify_url)
importFrom(httr,parse_url)
importFrom(httr,progress)
importFrom(httr,stop_for_status)
importFrom(httr,write_disk)
importFrom(methods,as)
importFrom(purrr,imap)
importFrom(purrr,keep)
importFrom(purrr,map)
importFrom(purrr,map_chr)
importFrom(purrr,map_dbl)
importFrom(purrr,map_int)
importFrom(purrr,pmap_chr)
importFrom(purrr,reduce)
importFrom(purrr,set_names)
importFrom(purrr,walk)
importFrom(rlang,.data)
importFrom(stats,setNames)
importFrom(stringr,str_remove_all)
importFrom(stringr,str_replace_all)
importFrom(tibble,column_to_rownames)
importFrom(tools,R_user_dir)
importFrom(utils,head)
importFrom(utils,packageName)

CuratedAtlasQueryR
================

<!-- badges: start -->

[![Lifecycle:maturing](https://img.shields.io/badge/lifecycle-maturing-blue.svg)](https://www.tidyverse.org/lifecycle/#maturing)
<!-- badges: end -->

`CuratedAtlasQuery` is a query interface that allow the programmatic
exploration and retrieval of the harmonised, curated and reannotated
CELLxGENE single-cell human cell atlas. Data can be retrieved at cell,
sample, or dataset levels based on filtering criteria.

Harmonised data is stored in the ARDC Nectar Research Cloud, and most
`CuratedAtlasQuery` functions interact with Nectar via web requests, so
a network connection is required for most functionality.

<img src="man/figures/logo.png" width="120x" height="139px" />

<img src="man/figures/svcf_logo.jpeg" width="155x" height="58px" /><img src="man/figures/czi_logo.png" width="129px" height="58px" /><img src="man/figures/bioconductor_logo.jpg" width="202px" height="58px" /><img src="man/figures/vca_logo.png" width="219px" height="58px" /><img src="man/figures/nectar_logo.png" width="180px" height="58px" />

# Query interface

## Installation

``` r
devtools::install_github("stemangiola/CuratedAtlasQueryR")
```

## Load the package

``` r
library(CuratedAtlasQueryR)
```

## Load and explore the metadata

### Load the metadata

``` r
# Note: in real applications you should use the default value of remote_url 
metadata <- get_metadata(remote_url = METADATA_URL)
metadata
#> # Source:   table</vast/scratch/users/milton.m/cache/R/CuratedAtlasQueryR/metadata.0.2.3.parquet> [?? x 56]
#> # Database: DuckDB 0.7.1 [unknown@Linux 3.10.0-1160.88.1.el7.x86_64:R 4.2.1/:memory:]
#>    cell_ sample_ cell_…¹ cell_…² confi…³ cell_…⁴ cell_…⁵ cell_…⁶ sampl…⁷ _samp…⁸
#>    <chr> <chr>   <chr>   <chr>     <dbl> <chr>   <chr>   <chr>   <chr>   <chr>  
#>  1 8387… 7bd7b8… natura… immune…       5 cd8 tem gmp     natura… 842ce7… Q59___…
#>  2 1768… 7bd7b8… natura… immune…       5 cd8 tem cd8 tcm natura… 842ce7… Q59___…
#>  3 6329… 7bd7b8… natura… immune…       5 cd8 tem clp     termin… 842ce7… Q59___…
#>  4 5027… 7bd7b8… natura… immune…       5 cd8 tem clp     natura… 842ce7… Q59___…
#>  5 7956… 7bd7b8… natura… immune…       5 cd8 tem clp     natura… 842ce7… Q59___…
#>  6 4305… 7bd7b8… natura… immune…       5 cd8 tem clp     termin… 842ce7… Q59___…
#>  7 2126… 933f96… natura… ilc           1 nk      nk      natura… c250bf… AML3__…
#>  8 3114… 933f96… natura… immune…       5 mait    nk      natura… c250bf… AML3__…
#>  9 1407… 933f96… natura… immune…       5 mait    clp     natura… c250bf… AML3__…
#> 10 2911… 933f96… natura… nk            5 nk      clp     natura… c250bf… AML3__…
#> # … with more rows, 46 more variables: assay <chr>,
#> #   assay_ontology_term_id <chr>, file_id_db <chr>,
#> #   cell_type_ontology_term_id <chr>, development_stage <chr>,
#> #   development_stage_ontology_term_id <chr>, disease <chr>,
#> #   disease_ontology_term_id <chr>, ethnicity <chr>,
#> #   ethnicity_ontology_term_id <chr>, experiment___ <chr>, file_id <chr>,
#> #   is_primary_data_x <chr>, organism <chr>, organism_ontology_term_id <chr>, …
```

The `metadata` variable can then be re-used for all subsequent queries.

### Explore the tissue

``` r
metadata |>
    dplyr::distinct(tissue, file_id) 
#> # Source:   SQL [10 x 2]
#> # Database: DuckDB 0.7.1 [unknown@Linux 3.10.0-1160.88.1.el7.x86_64:R 4.2.1/:memory:]
#>    tissue              file_id                             
#>    <chr>               <chr>                               
#>  1 bone marrow         1ff5cbda-4d41-4f50-8c7e-cbe4a90e38db
#>  2 lung parenchyma     6661ab3a-792a-4682-b58c-4afb98b2c016
#>  3 respiratory airway  6661ab3a-792a-4682-b58c-4afb98b2c016
#>  4 nose                6661ab3a-792a-4682-b58c-4afb98b2c016
#>  5 renal pelvis        dc9d8cdd-29ee-4c44-830c-6559cb3d0af6
#>  6 kidney              dc9d8cdd-29ee-4c44-830c-6559cb3d0af6
#>  7 renal medulla       dc9d8cdd-29ee-4c44-830c-6559cb3d0af6
#>  8 cortex of kidney    dc9d8cdd-29ee-4c44-830c-6559cb3d0af6
#>  9 kidney blood vessel dc9d8cdd-29ee-4c44-830c-6559cb3d0af6
#> 10 lung                a2796032-d015-40c4-b9db-835207e5bd5b
```

## Download single-cell RNA sequencing counts

### Query raw counts

``` r
single_cell_counts = 
    metadata |>
    dplyr::filter(
        ethnicity == "African" &
        stringr::str_like(assay, "%10x%") &
        tissue == "lung parenchyma" &
        stringr::str_like(cell_type, "%CD4%")
    ) |>
    get_single_cell_experiment()
#> ℹ Realising metadata.
#> ℹ Synchronising files
#> ℹ Downloading 0 files, totalling 0 GB
#> ℹ Reading files.
#> ℹ Compiling Single Cell Experiment.

single_cell_counts
#> # A SingleCellExperiment-tibble abstraction: 1,571 × 57
#> # [90mFeatures=36229 | Cells=1571 | Assays=counts[0m
#>    .cell sample_ cell_…¹ cell_…² confi…³ cell_…⁴ cell_…⁵ cell_…⁶ sampl…⁷ X_sam…⁸
#>    <chr> <chr>   <chr>   <chr>     <dbl> <chr>   <chr>   <chr>   <chr>   <chr>  
#>  1 AGCG… 11a7dc… CD4-po… cd4 th1       3 cd4 tcm cd8 t   th1     10b339… Donor_…
#>  2 TCAG… 11a7dc… CD4-po… cd4 th…       3 cd4 tcm cd4 tem th1/th… 10b339… Donor_…
#>  3 TTTA… 11a7dc… CD4-po… cd4 th…       3 cd4 tcm cd4 tcm th17    10b339… Donor_…
#>  4 ACAC… 11a7dc… CD4-po… immune…       5 cd4 tcm plasma  th1/th… 10b339… Donor_…
#>  5 CAAG… 11a7dc… CD4-po… immune…       1 cd4 tcm cd4 tcm mait    10b339… Donor_…
#>  6 CTGT… 14a078… CD4-po… cd4 th…       3 cd4 tcm cd4 tem th1/th… 8f71c5… VUHD85…
#>  7 ACGT… 14a078… CD4-po… treg          5 cd4 tcm tregs   t regu… 8f71c5… VUHD85…
#>  8 CATA… 14a078… CD4-po… immune…       5 nk      cd8 tem mait    8f71c5… VUHD85…
#>  9 ACTT… 14a078… CD4-po… mait          5 mait    cd8 tem mait    8f71c5… VUHD85…
#> 10 TGCG… 14a078… CD4-po… cd4 th1       3 cd4 tcm cd4 tem th1     8f71c5… VUHD85…
#> # … with 1,561 more rows, 47 more variables: assay <chr>,
#> #   assay_ontology_term_id <chr>, file_id_db <chr>,
#> #   cell_type_ontology_term_id <chr>, development_stage <chr>,
#> #   development_stage_ontology_term_id <chr>, disease <chr>,
#> #   disease_ontology_term_id <chr>, ethnicity <chr>,
#> #   ethnicity_ontology_term_id <chr>, experiment___ <chr>, file_id <chr>,
#> #   is_primary_data_x <chr>, organism <chr>, organism_ontology_term_id <chr>, …
```

### Query counts scaled per million

This is helpful if just few genes are of interest, as they can be
compared across samples.

``` r
single_cell_counts = 
    metadata |>
    dplyr::filter(
        ethnicity == "African" &
        stringr::str_like(assay, "%10x%") &
        tissue == "lung parenchyma" &
        stringr::str_like(cell_type, "%CD4%")
    ) |>
    get_single_cell_experiment(assays = "cpm")
#> ℹ Realising metadata.
#> ℹ Synchronising files
#> ℹ Downloading 0 files, totalling 0 GB
#> ℹ Reading files.
#> ℹ Compiling Single Cell Experiment.

single_cell_counts
#> # A SingleCellExperiment-tibble abstraction: 1,571 × 57
#> # [90mFeatures=36229 | Cells=1571 | Assays=cpm[0m
#>    .cell sample_ cell_…¹ cell_…² confi…³ cell_…⁴ cell_…⁵ cell_…⁶ sampl…⁷ X_sam…⁸
#>    <chr> <chr>   <chr>   <chr>     <dbl> <chr>   <chr>   <chr>   <chr>   <chr>  
#>  1 AGCG… 11a7dc… CD4-po… cd4 th1       3 cd4 tcm cd8 t   th1     10b339… Donor_…
#>  2 TCAG… 11a7dc… CD4-po… cd4 th…       3 cd4 tcm cd4 tem th1/th… 10b339… Donor_…
#>  3 TTTA… 11a7dc… CD4-po… cd4 th…       3 cd4 tcm cd4 tcm th17    10b339… Donor_…
#>  4 ACAC… 11a7dc… CD4-po… immune…       5 cd4 tcm plasma  th1/th… 10b339… Donor_…
#>  5 CAAG… 11a7dc… CD4-po… immune…       1 cd4 tcm cd4 tcm mait    10b339… Donor_…
#>  6 CTGT… 14a078… CD4-po… cd4 th…       3 cd4 tcm cd4 tem th1/th… 8f71c5… VUHD85…
#>  7 ACGT… 14a078… CD4-po… treg          5 cd4 tcm tregs   t regu… 8f71c5… VUHD85…
#>  8 CATA… 14a078… CD4-po… immune…       5 nk      cd8 tem mait    8f71c5… VUHD85…
#>  9 ACTT… 14a078… CD4-po… mait          5 mait    cd8 tem mait    8f71c5… VUHD85…
#> 10 TGCG… 14a078… CD4-po… cd4 th1       3 cd4 tcm cd4 tem th1     8f71c5… VUHD85…
#> # … with 1,561 more rows, 47 more variables: assay <chr>,
#> #   assay_ontology_term_id <chr>, file_id_db <chr>,
#> #   cell_type_ontology_term_id <chr>, development_stage <chr>,
#> #   development_stage_ontology_term_id <chr>, disease <chr>,
#> #   disease_ontology_term_id <chr>, ethnicity <chr>,
#> #   ethnicity_ontology_term_id <chr>, experiment___ <chr>, file_id <chr>,
#> #   is_primary_data_x <chr>, organism <chr>, organism_ontology_term_id <chr>, …
```

### Extract only a subset of genes

``` r
single_cell_counts = 
    metadata |>
    dplyr::filter(
        ethnicity == "African" &
        stringr::str_like(assay, "%10x%") &
        tissue == "lung parenchyma" &
        stringr::str_like(cell_type, "%CD4%")
    ) |>
    get_single_cell_experiment(assays = "cpm", features = "PUM1")
#> ℹ Realising metadata.
#> ℹ Synchronising files
#> ℹ Downloading 0 files, totalling 0 GB
#> ℹ Reading files.
#> ℹ Compiling Single Cell Experiment.

single_cell_counts
#> # A SingleCellExperiment-tibble abstraction: 1,571 × 57
#> # [90mFeatures=1 | Cells=1571 | Assays=cpm[0m
#>    .cell sample_ cell_…¹ cell_…² confi…³ cell_…⁴ cell_…⁵ cell_…⁶ sampl…⁷ X_sam…⁸
#>    <chr> <chr>   <chr>   <chr>     <dbl> <chr>   <chr>   <chr>   <chr>   <chr>  
#>  1 AGCG… 11a7dc… CD4-po… cd4 th1       3 cd4 tcm cd8 t   th1     10b339… Donor_…
#>  2 TCAG… 11a7dc… CD4-po… cd4 th…       3 cd4 tcm cd4 tem th1/th… 10b339… Donor_…
#>  3 TTTA… 11a7dc… CD4-po… cd4 th…       3 cd4 tcm cd4 tcm th17    10b339… Donor_…
#>  4 ACAC… 11a7dc… CD4-po… immune…       5 cd4 tcm plasma  th1/th… 10b339… Donor_…
#>  5 CAAG… 11a7dc… CD4-po… immune…       1 cd4 tcm cd4 tcm mait    10b339… Donor_…
#>  6 CTGT… 14a078… CD4-po… cd4 th…       3 cd4 tcm cd4 tem th1/th… 8f71c5… VUHD85…
#>  7 ACGT… 14a078… CD4-po… treg          5 cd4 tcm tregs   t regu… 8f71c5… VUHD85…
#>  8 CATA… 14a078… CD4-po… immune…       5 nk      cd8 tem mait    8f71c5… VUHD85…
#>  9 ACTT… 14a078… CD4-po… mait          5 mait    cd8 tem mait    8f71c5… VUHD85…
#> 10 TGCG… 14a078… CD4-po… cd4 th1       3 cd4 tcm cd4 tem th1     8f71c5… VUHD85…
#> # … with 1,561 more rows, 47 more variables: assay <chr>,
#> #   assay_ontology_term_id <chr>, file_id_db <chr>,
#> #   cell_type_ontology_term_id <chr>, development_stage <chr>,
#> #   development_stage_ontology_term_id <chr>, disease <chr>,
#> #   disease_ontology_term_id <chr>, ethnicity <chr>,
#> #   ethnicity_ontology_term_id <chr>, experiment___ <chr>, file_id <chr>,
#> #   is_primary_data_x <chr>, organism <chr>, organism_ontology_term_id <chr>, …
```

### Extract the counts as a Seurat object

This convert the H5 SingleCellExperiment to Seurat so it might take long
time and occupy a lot of memory depending on how many cells you are
requesting.

``` r
single_cell_counts_seurat = 
    metadata |>
    dplyr::filter(
        ethnicity == "African" &
        stringr::str_like(assay, "%10x%") &
        tissue == "lung parenchyma" &
        stringr::str_like(cell_type, "%CD4%")
    ) |>
    get_seurat()
#> ℹ Realising metadata.
#> ℹ Synchronising files
#> ℹ Downloading 0 files, totalling 0 GB
#> ℹ Reading files.
#> ℹ Compiling Single Cell Experiment.

single_cell_counts_seurat
#> An object of class Seurat 
#> 36229 features across 1571 samples within 1 assay 
#> Active assay: originalexp (36229 features, 0 variable features)
```

## Save your `SingleCellExperiment`

The returned `SingleCellExperiment` can be saved with two modalities, as
`.rds` or as `HDF5`.

### Saving as RDS (fast saving, slow reading)

Saving as `.rds` has the advantage of being fast, andd the `.rds` file
occupies very little disk space as it only stores the links to the files
in your cache.

However it has the disadvantage that for big `SingleCellExperiment`
objects, which merge a lot of HDF5 from your
`get_single_cell_experiment`, the display and manipulation is going to
be slow. In addition, an `.rds` saved in this way is not portable: you
will not be able to share it with other users.

``` r
single_cell_counts |> saveRDS("single_cell_counts.rds")
```

### Saving as HDF5 (slow saving, fast reading)

Saving as `.hdf5` executes any computation on the `SingleCellExperiment`
and writes it to disk as a monolithic `HDF5`. Once this is done,
operations on the `SingleCellExperiment` will be comparatively very
fast. The resulting `.hdf5` file will also be totally portable and
sharable.

However this `.hdf5` has the disadvantage of being larger than the
corresponding `.rds` as it includes a copy of the count information, and
the saving process is going to be slow for large objects.

``` r
single_cell_counts |> HDF5Array::saveHDF5SummarizedExperiment("single_cell_counts", replace = TRUE)
```

## Visualise gene transcription

We can gather all CD14 monocytes cells and plot the distribution of
HLA-A across all tissues

    #> ℹ Realising metadata.
    #> ℹ Synchronising files
    #> ℹ Downloading 0 files, totalling 0 GB
    #> ℹ Reading files.
    #> ℹ Compiling Single Cell Experiment.
    #> Warning: Transformation introduced infinite values in continuous y-axis
    #> Warning in min(x): no non-missing arguments to min; returning Inf
    #> Warning in max(x): no non-missing arguments to max; returning -Inf

![](/stornext/Home/data/allstaff/m/milton.m/HCAquery/README_files/figure-gfm/unnamed-chunk-14-1.png)<!-- -->

    #> Warning: Transformation introduced infinite values in continuous y-axis
    #> Warning in min(x): no non-missing arguments to min; returning Inf
    #> Warning in max(x): no non-missing arguments to max; returning -Inf

![](/stornext/Home/data/allstaff/m/milton.m/HCAquery/README_files/figure-gfm/unnamed-chunk-14-2.png)<!-- -->

``` r
library(tidySingleCellExperiment)
library(ggplot2)

counts |> 
  ggplot(aes( disease, `HLA.A`,color = file_id)) +
  geom_jitter(shape=".") 
```

![](/stornext/Home/data/allstaff/m/milton.m/HCAquery/README_files/figure-gfm/unnamed-chunk-15-1.png)<!-- -->

![](man/figures/HLA_A_disease_plot.png)<!-- -->

``` r
metadata |> 
    
  # Filter and subset
  dplyr::filter(cell_type_harmonised=="nk") |> 

  # Get counts per million for HCA-A gene 
  get_single_cell_experiment(assays = "cpm", features = "HLA-A") |> 

  # Plot (styling code have been omitted)
  tidySingleCellExperiment::join_features("HLA-A", shape = "wide") |> 
  ggplot(aes(tissue_harmonised, `HLA.A`,color = file_id)) +
  geom_jitter(shape=".")
#> ℹ Realising metadata.
#> ℹ Synchronising files
#> ℹ Downloading 0 files, totalling 0 GB
#> ℹ Reading files.
#> ℹ Compiling Single Cell Experiment.
```

![](/stornext/Home/data/allstaff/m/milton.m/HCAquery/README_files/figure-gfm/unnamed-chunk-17-1.png)<!-- -->

![](man/figures/HLA_A_tissue_plot.png)<!-- -->

## Obtain Unharmonised Metadata

Various metadata fields are *not* common between datasets, so it does
not make sense for these to live in the main metadata table. However, we
can obtain it using the `get_unharmonised_metadata()` function. This
function returns a data frame with one row per dataset, including the
`unharmonised` column which contains unharmnised metadata as a nested
data frame.

``` r
harmonised <- metadata |> dplyr::filter(tissue == "kidney blood vessel")
unharmonised <- get_unharmonised_metadata(harmonised)
unharmonised
#> # A tibble: 1 × 2
#>   file_id                              unharmonised   
#>   <chr>                                <list>         
#> 1 dc9d8cdd-29ee-4c44-830c-6559cb3d0af6 <tbl_dck_[,14]>
```

Notice that the columns differ between each dataset’s data frame:

``` r
dplyr::pull(unharmonised) |> head(2)
#> [[1]]
#> # Source:   SQL [?? x 14]
#> # Database: DuckDB 0.7.1 [unknown@Linux 3.10.0-1160.88.1.el7.x86_64:R 4.2.1/:memory:]
#>    cell_ file_id donor…¹ donor…² libra…³ mappe…⁴ sampl…⁵ suspe…⁶ suspe…⁷ autho…⁸
#>    <chr> <chr>   <chr>   <chr>   <chr>   <chr>   <chr>   <chr>   <chr>   <chr>  
#>  1 4602… dc9d8c… 27 mon… a8536b… 5ddaea… GENCOD… 61bf84… cell    d8a44f… Pelvic…
#>  2 4602… dc9d8c… 27 mon… a8536b… 5ddaea… GENCOD… 61bf84… cell    d8a44f… Pelvic…
#>  3 4602… dc9d8c… 27 mon… a8536b… 5ddaea… GENCOD… 61bf84… cell    d8a44f… Pelvic…
#>  4 4602… dc9d8c… 27 mon… a8536b… 5ddaea… GENCOD… 61bf84… cell    d8a44f… Pelvic…
#>  5 4602… dc9d8c… 27 mon… a8536b… 5ddaea… GENCOD… 61bf84… cell    d8a44f… Pelvic…
#>  6 4602… dc9d8c… 27 mon… a8536b… 5ddaea… GENCOD… 61bf84… cell    d8a44f… Pelvic…
#>  7 4602… dc9d8c… 27 mon… a8536b… 5ddaea… GENCOD… 61bf84… cell    d8a44f… Pelvic…
#>  8 4602… dc9d8c… 27 mon… a8536b… 5ddaea… GENCOD… 61bf84… cell    d8a44f… Pelvic…
#>  9 4602… dc9d8c… 27 mon… a8536b… 5ddaea… GENCOD… 61bf84… cell    d8a44f… Pelvic…
#> 10 4602… dc9d8c… 27 mon… a8536b… 5ddaea… GENCOD… 61bf84… cell    d8a44f… Pelvic…
#> # … with more rows, 4 more variables: reported_diseases <chr>,
#> #   Experiment <chr>, Project <chr>, broad_celltype <chr>, and abbreviated
#> #   variable names ¹​donor_age, ²​donor_uuid, ³​library_uuid,
#> #   ⁴​mapped_reference_annotation, ⁵​sample_uuid, ⁶​suspension_type,
#> #   ⁷​suspension_uuid, ⁸​author_cell_type
```

# Cell metadata

Dataset-specific columns (definitions available at
cellxgene.cziscience.com)

`cell_count`, `collection_id`, `created_at.x`, `created_at.y`,
`dataset_deployments`, `dataset_id`, `file_id`, `filename`, `filetype`,
`is_primary_data.y`, `is_valid`, `linked_genesets`,
`mean_genes_per_cell`, `name`, `published`, `published_at`,
`revised_at`, `revision`, `s3_uri`, `schema_version`, `tombstone`,
`updated_at.x`, `updated_at.y`, `user_submitted`, `x_normalization`

Sample-specific columns (definitions available at
cellxgene.cziscience.com)

`sample_`, `sample_name`, `age_days`, `assay`, `assay_ontology_term_id`,
`development_stage`, `development_stage_ontology_term_id`, `ethnicity`,
`ethnicity_ontology_term_id`, `experiment___`, `organism`,
`organism_ontology_term_id`, `sample_placeholder`, `sex`,
`sex_ontology_term_id`, `tissue`, `tissue_harmonised`,
`tissue_ontology_term_id`, `disease`, `disease_ontology_term_id`,
`is_primary_data.x`

Cell-specific columns (definitions available at
cellxgene.cziscience.com)

`cell_`, `cell_type`, `cell_type_ontology_term_idm`,
`cell_type_harmonised`, `confidence_class`,
`cell_annotation_azimuth_l2`, `cell_annotation_blueprint_singler`

Through harmonisation and curation we introduced custom column, not
present in the original CELLxGENE metadata

- `tissue_harmonised`: a coarser tissue name for better filtering
- `age_days`: the number of days corresponding to the age
- `cell_type_harmonised`: the consensus call identity (for immune cells)
  using the original and three novel annotations using Seurat Azimuth
  and SingleR
- `confidence_class`: an ordinal class of how confident
  `cell_type_harmonised` is. 1 is complete consensus, 2 is 3 out of four
  and so on.  
- `cell_annotation_azimuth_l2`: Azimuth cell annotation
- `cell_annotation_blueprint_singler`: SingleR cell annotation using
  Blueprint reference
- `cell_annotation_blueprint_monaco`: SingleR cell annotation using
  Monaco reference
- `sample_id_db`: Sample subdivision for internal use
- `file_id_db`: File subdivision for internal use
- `sample_`: Sample ID
- `.sample_name`: How samples were defined

# RNA abundance

The `raw` assay includes RNA abundance in the positive real scale (not
transformed with non-linear functions, e.g. log sqrt). Originally
CELLxGENE include a mix of scales and transformations specified in the
`x_normalization` column.

The `cpm` assay includes counts per million.

# Session Info

``` r
sessionInfo()
#> R version 4.2.1 (2022-06-23)
#> Platform: x86_64-pc-linux-gnu (64-bit)
#> Running under: CentOS Linux 7 (Core)
#> 
#> Matrix products: default
#> BLAS:   /stornext/System/data/apps/R/R-4.2.1/lib64/R/lib/libRblas.so
#> LAPACK: /stornext/System/data/apps/R/R-4.2.1/lib64/R/lib/libRlapack.so
#> 
#> locale:
#>  [1] LC_CTYPE=en_US.UTF-8       LC_NUMERIC=C              
#>  [3] LC_TIME=en_US.UTF-8        LC_COLLATE=en_US.UTF-8    
#>  [5] LC_MONETARY=en_US.UTF-8    LC_MESSAGES=en_US.UTF-8   
#>  [7] LC_PAPER=en_US.UTF-8       LC_NAME=C                 
#>  [9] LC_ADDRESS=C               LC_TELEPHONE=C            
#> [11] LC_MEASUREMENT=en_US.UTF-8 LC_IDENTIFICATION=C       
#> 
#> attached base packages:
#> [1] stats4    stats     graphics  grDevices utils     datasets  methods  
#> [8] base     
#> 
#> other attached packages:
#>  [1] tidySingleCellExperiment_1.6.3 SingleCellExperiment_1.18.1   
#>  [3] SummarizedExperiment_1.26.1    Biobase_2.56.0                
#>  [5] GenomicRanges_1.48.0           GenomeInfoDb_1.32.4           
#>  [7] IRanges_2.30.1                 S4Vectors_0.34.0              
#>  [9] BiocGenerics_0.42.0            MatrixGenerics_1.8.1          
#> [11] matrixStats_0.63.0             ttservice_0.2.2               
#> [13] ggplot2_3.4.1                  CuratedAtlasQueryR_0.99.1     
#> 
#> loaded via a namespace (and not attached):
#>   [1] plyr_1.8.8             igraph_1.4.1           lazyeval_0.2.2        
#>   [4] sp_1.5-1               splines_4.2.1          listenv_0.9.0         
#>   [7] scattermore_0.8        digest_0.6.31          htmltools_0.5.4       
#>  [10] fansi_1.0.3            magrittr_2.0.3         tensor_1.5            
#>  [13] cluster_2.1.3          ROCR_1.0-11            globals_0.16.2        
#>  [16] duckdb_0.7.1-1         spatstat.sparse_3.0-0  colorspace_2.0-3      
#>  [19] blob_1.2.3             ggrepel_0.9.2          xfun_0.36             
#>  [22] dplyr_1.1.0            RCurl_1.98-1.9         jsonlite_1.8.4        
#>  [25] progressr_0.13.0       spatstat.data_3.0-0    survival_3.3-1        
#>  [28] zoo_1.8-11             glue_1.6.2             polyclip_1.10-4       
#>  [31] gtable_0.3.1           zlibbioc_1.42.0        XVector_0.36.0        
#>  [34] leiden_0.4.3           DelayedArray_0.22.0    Rhdf5lib_1.18.2       
#>  [37] future.apply_1.10.0    HDF5Array_1.24.2       abind_1.4-5           
#>  [40] scales_1.2.1           DBI_1.1.3              spatstat.random_3.0-1 
#>  [43] miniUI_0.1.1.1         Rcpp_1.0.10            viridisLite_0.4.1     
#>  [46] xtable_1.8-4           reticulate_1.26        htmlwidgets_1.6.0     
#>  [49] httr_1.4.4             RColorBrewer_1.1-3     ellipsis_0.3.2        
#>  [52] Seurat_4.3.0           ica_1.0-3              farver_2.1.1          
#>  [55] pkgconfig_2.0.3        dbplyr_2.3.0           sass_0.4.4            
#>  [58] uwot_0.1.14            deldir_1.0-6           utf8_1.2.2            
#>  [61] labeling_0.4.2         tidyselect_1.2.0       rlang_1.0.6           
#>  [64] reshape2_1.4.4         later_1.3.0            munsell_0.5.0         
#>  [67] tools_4.2.1            cachem_1.0.6           cli_3.6.0             
#>  [70] generics_0.1.3         ggridges_0.5.4         evaluate_0.19         
#>  [73] stringr_1.5.0          fastmap_1.1.0          yaml_2.3.6            
#>  [76] goftest_1.2-3          knitr_1.42             fitdistrplus_1.1-8    
#>  [79] purrr_1.0.1            RANN_2.6.1             pbapply_1.6-0         
#>  [82] future_1.30.0          nlme_3.1-157           mime_0.12             
#>  [85] compiler_4.2.1         rstudioapi_0.14        curl_4.3.3            
#>  [88] plotly_4.10.1          png_0.1-8              spatstat.utils_3.0-1  
#>  [91] tibble_3.1.8           bslib_0.4.2            stringi_1.7.12        
#>  [94] highr_0.10             forcats_1.0.0          lattice_0.20-45       
#>  [97] Matrix_1.5-3           vctrs_0.5.2            pillar_1.8.1          
#> [100] lifecycle_1.0.3        rhdf5filters_1.8.0     spatstat.geom_3.0-3   
#> [103] lmtest_0.9-40          jquerylib_0.1.4        RcppAnnoy_0.0.20      
#> [106] data.table_1.14.6      cowplot_1.1.1          bitops_1.0-7          
#> [109] irlba_2.3.5.1          httpuv_1.6.7           patchwork_1.1.2       
#> [112] R6_2.5.1               promises_1.2.0.1       KernSmooth_2.23-20    
#> [115] gridExtra_2.3          parallelly_1.33.0      codetools_0.2-18      
#> [118] assertthat_0.2.1       MASS_7.3-57            rhdf5_2.40.0          
#> [121] rprojroot_2.0.3        withr_2.5.0            SeuratObject_4.1.3    
#> [124] sctransform_0.3.5      GenomeInfoDbData_1.2.8 parallel_4.2.1        
#> [127] grid_4.2.1             tidyr_1.3.0            rmarkdown_2.20        
#> [130] Rtsne_0.16             spatstat.explore_3.0-5 shiny_1.7.4
```

params <-
list(demo_metadata = TRUE)

## ----include = FALSE----------------------------------------------------------
knitr::opts_chunk$set(
    collapse = TRUE,
    comment = "#>"
)
root_dir <- knitr::opts_knit$get("root.dir")
if (!is.null(root_dir)){
    # This hack fixes the relative image paths. 
    # See https://github.com/rstudio/rmarkdown/issues/2473
    knitr::opts_knit$set(
        output.dir = root_dir
    )
}
proj_root <- rprojroot::find_package_root_file() |> normalizePath() 
# Utility function for figures to force them to have the correct path
find_figure <- function(names){
    rprojroot::find_package_root_file() |>
        file.path("man", "figures", names) 
}
METADATA_URL = if (params$demo_metadata)
        CuratedAtlasQueryR::SAMPLE_DATABASE_URL else
        CuratedAtlasQueryR::DATABASE_URL

## ----echo=FALSE, out.height = c("139px"), out.width = "120x"------------------
find_figure("logo.png") |> knitr::include_graphics()

## ----echo=FALSE, out.height = c("58px"), out.width = c("155x", "129px", "202px", "219px", "180px")----
c(
  "svcf_logo.jpeg", 
  "czi_logo.png", 
  "bioconductor_logo.jpg",
  "vca_logo.png",
  "nectar_logo.png"
) |> 
    find_figure() |>
    knitr::include_graphics()

## ----eval=FALSE---------------------------------------------------------------
# devtools::install_github("stemangiola/CuratedAtlasQueryR")

## ----message=FALSE, warning=FALSE---------------------------------------------
library(CuratedAtlasQueryR)

## -----------------------------------------------------------------------------
# Note: in real applications you should use the default value of remote_url 
metadata <- get_metadata(remote_url = METADATA_URL)
metadata

## -----------------------------------------------------------------------------
metadata |>
    dplyr::distinct(tissue, file_id) 

## -----------------------------------------------------------------------------
single_cell_counts = 
    metadata |>
    dplyr::filter(
        ethnicity == "African" &
        stringr::str_like(assay, "%10x%") &
        tissue == "lung parenchyma" &
        stringr::str_like(cell_type, "%CD4%")
    ) |>
    get_single_cell_experiment()

single_cell_counts

## -----------------------------------------------------------------------------
single_cell_counts = 
    metadata |>
    dplyr::filter(
        ethnicity == "African" &
        stringr::str_like(assay, "%10x%") &
        tissue == "lung parenchyma" &
        stringr::str_like(cell_type, "%CD4%")
    ) |>
    get_single_cell_experiment(assays = "cpm")

single_cell_counts

## -----------------------------------------------------------------------------
single_cell_counts = 
    metadata |>
    dplyr::filter(
        ethnicity == "African" &
        stringr::str_like(assay, "%10x%") &
        tissue == "lung parenchyma" &
        stringr::str_like(cell_type, "%CD4%")
    ) |>
    get_single_cell_experiment(assays = "cpm", features = "PUM1")

single_cell_counts

## -----------------------------------------------------------------------------
single_cell_counts_seurat = 
    metadata |>
    dplyr::filter(
        ethnicity == "African" &
        stringr::str_like(assay, "%10x%") &
        tissue == "lung parenchyma" &
        stringr::str_like(cell_type, "%CD4%")
    ) |>
    get_seurat()

single_cell_counts_seurat

## -----------------------------------------------------------------------------
single_cell_counts |> saveRDS("single_cell_counts.rds")

## -----------------------------------------------------------------------------
single_cell_counts |> HDF5Array::saveHDF5SummarizedExperiment("single_cell_counts", replace = TRUE)

## ----results='hide'-----------------------------------------------------------
suppressPackageStartupMessages({
    library(ggplot2)
})

# Plots with styling
counts <- metadata |>
  # Filter and subset
  dplyr::filter(cell_type_harmonised == "cd14 mono") |>
  dplyr::filter(file_id_db != "c5a05f23f9784a3be3bfa651198a48eb") |> 
  
  # Get counts per million for HCA-A gene
  get_single_cell_experiment(assays = "cpm", features = "HLA-A") |> 
  suppressMessages() |>
  
  # Add feature to table
  tidySingleCellExperiment::join_features("HLA-A", shape = "wide") |> 
    
  # Rank x axis
  tibble::as_tibble()

# Plot by disease
counts |>
  dplyr::with_groups(disease, ~ .x |> dplyr::mutate(median_count = median(`HLA.A`, rm.na=TRUE))) |> 
  
  # Plot
  ggplot(aes(forcats::fct_reorder(disease, median_count,.desc = TRUE), `HLA.A`,color = file_id)) +
  geom_jitter(shape=".") +
    
  # Style
  guides(color="none") +
  scale_y_log10() +
  theme_bw() +
  theme(axis.text.x = element_text(angle = 60, vjust = 1, hjust = 1)) + 
  xlab("Disease") + 
  ggtitle("HLA-A in CD14 monocytes by disease") 

## ----echo=FALSE---------------------------------------------------------------
find_figure("HLA_A_disease_plot.png") |> knitr::include_graphics()

## ----results='hide'-----------------------------------------------------------
# Plot by tissue
counts |> 
  dplyr::with_groups(tissue_harmonised, ~ .x |> dplyr::mutate(median_count = median(`HLA.A`, rm.na=TRUE))) |> 
  
  # Plot
  ggplot(aes(forcats::fct_reorder(tissue_harmonised, median_count,.desc = TRUE), `HLA.A`,color = file_id)) +
  geom_jitter(shape=".") +
    
  # Style
  guides(color="none") +
  scale_y_log10() +
  theme_bw() +
  theme(axis.text.x = element_text(angle = 60, vjust = 1, hjust = 1)) + 
  xlab("Tissue") + 
  ggtitle("HLA-A in CD14 monocytes by tissue") + 
  theme(legend.position = "none")

## ----echo=FALSE---------------------------------------------------------------
find_figure("HLA_A_tissue_plot.png") |> knitr::include_graphics()

## -----------------------------------------------------------------------------
metadata |> 
    
  # Filter and subset
  dplyr::filter(cell_type_harmonised=="nk") |> 

  # Get counts per million for HCA-A gene 
  get_single_cell_experiment(assays = "cpm", features = "HLA-A") |> 
  suppressMessages() |>

  # Plot
  tidySingleCellExperiment::join_features("HLA-A", shape = "wide") |> 
  ggplot(aes(tissue_harmonised, `HLA.A`, color = file_id)) +
  theme_bw() +
  theme(
      axis.text.x = element_text(angle = 60, vjust = 1, hjust = 1),
      legend.position = "none"
  ) + 
  geom_jitter(shape=".") + 
  xlab("Tissue") + 
  ggtitle("HLA-A in nk cells by tissue")

## -----------------------------------------------------------------------------
harmonised <- metadata |> dplyr::filter(tissue == "kidney blood vessel")
unharmonised <- get_unharmonised_metadata(harmonised)
unharmonised

## -----------------------------------------------------------------------------
dplyr::pull(unharmonised) |> head(2)

## -----------------------------------------------------------------------------
sessionInfo()

if(requireNamespace('spelling', quietly = TRUE))
  spelling::spell_check_test(
    # We hit https://github.com/ropensci/spelling/issues/62 if the 
    # vignettes are checked
    vignettes = FALSE,
    error = FALSE,
    skip_on_cran = TRUE
 )

#' @importFrom purrr walk
#' @importFrom dbplyr remote_con
#' @importFrom DBI dbDisconnect
.onUnload <- function(libname, pkgname){
    # Close connections to all cached tables. This should avoid most of the
    # "Connection is garbage-collected" messages
    cache$metadata_table |>
        as.list() |>
        walk(function(table){
            table |>
                remote_con() |>
                dbDisconnect()
        })
}
.onAttach <- function(libname, pkgname) {
    msg <- sprintf(
        "Package '%s' is deprecated and will be removed from Bioconductor
         version %s", pkgname, "3.24")
    .Deprecated(msg=paste(strwrap(msg, exdent=2), collapse="\n"))
}

# Functions that relate to the Seurat conversion

#' @importFrom assertthat assert_that
#' @importFrom methods as
#' @importFrom SeuratObject as.sparse
#' @exportS3Method
as.sparse.DelayedMatrix <- function(x, ...) {
    # This is glue to ensure the SCE -> Seurat conversion works properly with
    # DelayedArray types
    as(x, "dgCMatrix")
}

#' Given a data frame of HCA metadata, returns a Seurat object corresponding to
#' the samples in that data frame
#'
#' @inheritDotParams get_single_cell_experiment
#' @importFrom SeuratObject as.Seurat
#' @export
#' @return A Seurat object containing the same data as a call to
#'   [get_single_cell_experiment()]
#' @examples
#' meta <- get_metadata() |> head(2)
#' seurat <- get_seurat(meta)
#'
get_seurat <- function(...) {
    get_single_cell_experiment(...) |> as.Seurat(data = NULL)
}

# Functions that relate to downloading count data into SingleCellExperiments

# We need to load utils now so it can be used at the top level
#' @include utils.R
# This is a hack to force Seurat packages to be loaded, and also to
# satisfy R CMD check. We don't need to attach them at all.
#' @importFrom Seurat as.SingleCellExperiment
NULL

# Maps user provided assay names to their corresponding paths in the repository
assay_map <- c(
    counts = "original",
    cpm = "cpm"
)

#' Base URL pointing to the count data at the current version
#' @noRd
COUNTS_URL <- single_line_str(
    "https://object-store.rc.nectar.org.au/v1/
    AUTH_06d6e008e3e642da99d806ba3ea629c5/cellxgene-0.2.1-hdf5"
)
#' Current version of the counts. This will be incremented when a newer
#' version is released
#' @noRd
COUNTS_VERSION <- "0.2.1"

#' @inherit get_single_cell_experiment
#' @inheritDotParams get_single_cell_experiment
#' @importFrom cli cli_alert_warning
#' @export
get_SingleCellExperiment <- function(...){
    single_line_str("This function name is deprecated. 
                    Please use `get_single_cell_experiment()` instead") |>
        cli_alert_warning()
    
    get_single_cell_experiment(...)
}

#' Gets a SingleCellExperiment from curated metadata
#'
#' Given a data frame of Curated Atlas metadata obtained from [get_metadata()],
#' returns a [`SingleCellExperiment::SingleCellExperiment-class`] object
#' corresponding to the samples in that data frame
#'
#' @param data A data frame containing, at minimum, a `sample_` column, which
#'   corresponds to a single cell sample ID. This can be obtained from the
#'   [get_metadata()] function.
#' @param assays A character vector whose elements must be either "counts"
#'   and/or "cpm", representing the corresponding assay(s) you want to request.
#'   By default only the count assay is downloaded. If you are interested in
#'   comparing a limited amount of genes, the "cpm" assay is more appropriate.
#' @param repository A character vector of length one. If provided, it should be
#'   an HTTP URL pointing to the location where the single cell data is stored.
#' @param cache_directory An optional character vector of length one. If
#'   provided, it should indicate a local file path where any remotely accessed
#'   files should be copied.
#' @param features An optional character vector of features (ie genes) to return
#'   the counts for. By default counts for all features will be returned.
#' @returns A SingleCellExperiment object, with one assay for each value in the
#'   assays argument
#' @examples
#' meta <- get_metadata() |> head(2)
#' sce <- get_single_cell_experiment(meta)
#'
#' @importFrom dplyr pull filter as_tibble inner_join collect
#' @importFrom tibble column_to_rownames
#' @importFrom purrr reduce map map_int imap keep
#' @importFrom BiocGenerics cbind
#' @importFrom glue glue
#' @importFrom HDF5Array loadHDF5SummarizedExperiment HDF5RealizationSink
#'   loadHDF5SummarizedExperiment
#' @importFrom SingleCellExperiment SingleCellExperiment combineCols
#' @importFrom SummarizedExperiment colData assayNames<-
#' @importFrom httr parse_url
#' @importFrom assertthat assert_that has_name
#' @importFrom cli cli_alert_success cli_alert_info
#' @importFrom rlang .data
#' @importFrom stats setNames
#' @importFrom S4Vectors DataFrame
#' @export
get_single_cell_experiment <- function(
    data,
    assays = "counts",
    cache_directory = get_default_cache_dir(),
    repository = COUNTS_URL,
    features = NULL
) {
    # Parameter validation
    
    assays %in% names(assay_map) |>
        all() |>
        assert_that(
            msg = 'assays must be a character vector containing "counts" and/or
            "cpm"'
        )
    assert_that(
        !anyDuplicated(assays),
        inherits(cache_directory, "character"),
        is.null(repository) || is.character(repository),
        is.null(features) || is.character(features)
    )

    # Data parameter validation (last, because it's slower)
    ## Evaluate the promise now so that we get a sensible error message
    force(data)
    ## We have to convert to an in-memory table here, or some of the dplyr
    ## operations will fail when passed a database connection
    cli_alert_info("Realising metadata.")
    raw_data <- collect(data)
    assert_that(
        inherits(raw_data, "tbl"),
        has_name(raw_data, c("cell_", "file_id_db"))
    )

    versioned_cache_directory <- file.path(cache_directory, COUNTS_VERSION)
    versioned_cache_directory |> dir.create(
        showWarnings = FALSE,
        recursive = TRUE
    )

    subdirs <- assay_map[assays]

    # The repository is optional. If not provided we load only from the cache
    if (!is.null(repository)) {
        cli_alert_info("Synchronising files")
        parsed_repo <- parse_url(repository)
        parsed_repo$scheme |>
            `%in%`(c("http", "https")) |>
            assert_that()

        files_to_read <-
            raw_data |>
            pull(.data$file_id_db) |>
            unique() |>
            as.character() |>
            sync_assay_files(
                url = parsed_repo,
                cache_dir = versioned_cache_directory,
                files = _,
                subdirs = subdirs
            )
    }

    cli_alert_info("Reading files.")
    sces <- subdirs |>
        imap(function(current_subdir, current_assay) {
            # Build up an SCE for each assay
            dir_prefix <- file.path(
                versioned_cache_directory,
                current_subdir
            )

            raw_data |>
                dplyr::group_by(.data$file_id_db) |>
                # Load each file and attach metadata
                dplyr::summarise(sces = list(group_to_sce(
                    dplyr::cur_group_id(),
                    dplyr::cur_data_all(),
                    dir_prefix,
                    features
                ))) |>
                dplyr::pull(sces) |>
                # Combine each sce by column, since each sce has a different set
                # of cells
                do.call(cbind, args = _)
        })

    cli_alert_info("Compiling Single Cell Experiment.")
    # Combine all the assays
    sce <- sces[[1]]
    SummarizedExperiment::assays(sce) <- map(sces, function(sce) {
        SummarizedExperiment::assays(sce)[[1]]
    })

    sce
}

#' Converts a data frame into a single SCE
#' @param i Suffix to be added to the column names, to make them unique
#' @param df The data frame to be converted
#' @param dir_prefix The path to the single cell experiment, minus the final
#'   segment
#' @param features The list of genes/rows of interest
#' @return A SingleCellExperiment object
#' @importFrom dplyr mutate filter
#' @importFrom HDF5Array loadHDF5SummarizedExperiment
#' @importFrom SummarizedExperiment colData<-
#' @importFrom tibble column_to_rownames
#' @importFrom utils head
#' @importFrom cli cli_alert_warning cli_abort
#' @importFrom glue glue
#' @importFrom stringr str_replace_all
#' @noRd
group_to_sce <- function(i, df, dir_prefix, features) {
    sce_path <- df$file_id_db |>
        head(1) |>
        file.path(
            dir_prefix,
            suffix = _
        )

    file.exists(sce_path) |>
        assert_that(
            msg = "Your cache does not contain a file {sce_path} you
                            attempted to query. Please provide the repository
                            parameter so that files can be synchronised from the
                            internet" |> glue()
        )

    sce <- loadHDF5SummarizedExperiment(sce_path)
    # The cells we select here are those that are both available in the SCE
    # object, and requested for this particular file
    cells <- colnames(sce) |> intersect(df$cell_)

    if (length(cells) < nrow(df)){
        single_line_str(
            "Some cells were filtered out while loading {head(df$file_id_db, 1)}
            because of extremely low counts. The
            number of cells in the SingleCellExperiment will be less than the
            number of cells you have selected from the metadata."
        ) |> cli_alert_warning()
        df <- filter(df, .data$cell_ %in% cells)
    }
    else if (length(cells) > nrow(df)){
        cli_abort("This should never happen")
    }
    
    # Fix for https://github.com/tidyverse/dplyr/issues/6746
    force(i)
    
    new_coldata <- df |>
        # We need to make the cell names globally unique, which we can guarantee
        # by adding a suffix that is derived from file_id_db, which is the
        # grouping variable
        mutate(original_cell_id = .data$cell_, cell_ = glue("{cell_}_{i}")) |>
        column_to_rownames("cell_") |>
        as("DataFrame")
    
    `if`(
            is.null(features),
            sce[, new_coldata$original_cell_id],
            {
                # Optionally subset the genes
                genes <- rownames(sce) |> intersect(features)
                sce[genes, new_coldata$original_cell_id]
            }
    ) |>
        `colnames<-`(new_coldata$cell_) |>
        `colData<-`(value = new_coldata)
}

#' Synchronises one or more remote assays with a local copy
#' @param url A character vector of length one. The base HTTP URL from which to
#'   obtain the files.
#' @param cache_dir A character vector of length one. The local filepath to
#'   synchronise files to.
#' @param subdirs A character vector of subdirectories within the root URL to
#'   sync. These correspond to assays.
#' @param files A character vector containing one or more file_id_db entries
#' @returns A character vector consisting of file paths to all the newly
#'   downloaded files
#' @return A character vector of files that have been downloaded
#' @importFrom purrr pmap_chr map_chr
#' @importFrom httr modify_url
#' @importFrom dplyr transmute filter
#' @noRd
#'
sync_assay_files <- function(
    url = parse_url(COUNTS_URL),
    cache_dir,
    subdirs,
    files
) {
    # Find every combination of file name, sample id, and assay, since each
    # will be a separate file we need to download
    files <- expand.grid(
        filename = c("assays.h5", "se.rds"),
        sample_id = files,
        subdir = subdirs,
        stringsAsFactors = FALSE
    ) |>
        transmute(
            # Path to the file of interest from the root path. We use "/"
            # since URLs must use these regardless of OS
            full_url = paste0(
                url$path,
                "/",
                .data$subdir,
                "/",
                .data$sample_id,
                "/",
                .data$filename
            ) |> map_chr(~ modify_url(url, path = .)),

            # Path to save the file on local disk (and its parent directory)
            # We use file.path since the file separator will differ on other OSs
            output_dir = file.path(
                cache_dir,
                .data$subdir,
                .data$sample_id
            ),
            output_file = file.path(
                .data$output_dir,
                .data$filename
            )
        ) |>
        filter(
            # Don't bother downloading files that don't exist TODO: use some
            # kind of hashing to check if the remote file has changed, and
            # proceed with the download if it has. However this is low
            # importance as the repository is not likely to change often
            !file.exists(.data$output_file)
        )
    
    report_file_sizes(files$full_url)
    
    pmap_chr(files, function(full_url, output_dir, output_file) {
        sync_remote_file(full_url, output_file)
        output_file
    }, .progress = list(name = "Downloading files"))
}

library(dplyr)

test_that("get_SingleCellExperiment() correctly handles duplicate cell IDs", {
    meta <- get_metadata() |>
        dplyr::filter(cell_ == "868417_1") |>
        dplyr::collect()
    sce <- get_SingleCellExperiment(meta)
    # This query should return multiple cells, despite querying only 1 cell ID
    nrow(meta) |> expect_gt(1)
    # Each of the two ambiguous cell IDs should now be unique
    colnames(sce) |> expect_equal(c("868417_1_1", "868417_1_2"))
    # We should have lots of column data, derived from the metadata
    SummarizedExperiment::colData(sce) |>
        dim() |>
        expect_equal(c(2, 56))
})

test_that("get_default_cache_dir() returns the correct directory on Linux", {
    grepl("linux", version$platform, fixed = TRUE) |>
        skip_if_not()

    "~/.cache/R/CuratedAtlasQueryR" |>
        normalizePath() |>
        expect_equal(
            get_default_cache_dir(),
        )
})

test_that("sync_assay_files() syncs appropriate files", {
    temp <- tempfile()
    test_file <- "00095cb0de0dc9528316b636fc9b3446"

    sync_assay_files(
        subdirs = "cpm",
        cache_dir = temp,
        files = test_file
    )

    temp_subdir <- file.path(temp, "cpm")

    test_file %in% list.files(temp_subdir) |>
        expect(failure_message = "The correct subdirectory was not created")

    downloaded <- file.path(temp_subdir, test_file) |>
        list.files()

    expect_equal(downloaded, c("assays.h5", "se.rds"))
})

test_that("get_SingleCellExperiment() syncs appropriate files", {
    temp <- tempfile()
    test_file <- "00095cb0de0dc9528316b636fc9b3446"

    meta <- get_metadata() |> head(2)

    # The remote dataset should have many genes
    sce <- get_SingleCellExperiment(meta, cache_directory = temp)
    sce |>
        row.names() |>
        length() |>
        expect_gt(1)
})

test_that(
    "The assays argument to get_SingleCellExperiment controls the number
  of returned assays",
    {
        # We need this for the assays() function
        library(SummarizedExperiment)

        meta <- get_metadata() |> head(2)

        # If we request both assays, we get both assays
        get_SingleCellExperiment(meta, assays = c("counts", "cpm")) |>
            assays() |>
            names() |>
            expect_setequal(c("counts", "cpm"))

        # If we request one assay, we get one assays
        get_SingleCellExperiment(meta, assays = "counts") |>
            assays() |>
            names() |>
            expect_setequal("counts")
        get_SingleCellExperiment(meta, assays = "cpm") |>
            assays() |>
            names() |>
            expect_setequal("cpm")
    }
)

test_that("The features argument to get_SingleCellExperiment subsets genes", {
    meta <- get_metadata() |> head(2)

    # The un-subset dataset should have many genes
    sce_full <- get_SingleCellExperiment(meta) |>
        row.names() |>
        length()
    expect_gt(sce_full, 1)

    # The subset dataset should only have one gene
    sce_subset <- get_SingleCellExperiment(meta, features = "PUM1") |>
        row.names() |>
        length()
    expect_equal(sce_subset, 1)

    expect_gt(sce_full, sce_subset)
})

test_that("get_seurat() returns the appropriate data in Seurat format", {
    meta <- get_metadata() |> head(2)

    sce <- get_SingleCellExperiment(meta, features = "PUM1")
    seurat <- get_seurat(meta, features = "PUM1")

    # The output should be a Seurat object
    expect_s4_class(seurat, "Seurat")
    # Both methods should have appropriately subset genes
    expect_equal(
        rownames(sce),
        rownames(seurat)
    )
})

test_that("get_SingleCellExperiment() assigns the right cell ID to each cell", {
    id = "3214d8f8986c1e33a85be5322f2db4a9"
    
    # Force the file to be cached
    get_metadata() |>
        filter(file_id_db == id) |>
        get_SingleCellExperiment()
    
    # Load the SCE from cache directly
    assay_1 = CuratedAtlasQueryR:::get_default_cache_dir() |>
        file.path(CuratedAtlasQueryR:::COUNTS_VERSION, "original", id) |>
        HDF5Array::loadHDF5SummarizedExperiment() |>
        assay("X") |>
        as.matrix()
    
    # Make a SCE that has the right column names, but reversed
    assay_2 = 
        assay_1 |>
        colnames() |>
        tibble::tibble(
            file_id_db = id,
            cell_ = _
        ) |>
        arrange(-row_number()) |>
        get_SingleCellExperiment(assays = "counts") |>
        assay("counts") |>
        as.matrix()
        
    colnames(assay_2) = sub("_1", "", x=colnames(assay_2))

    expect_equal(
        assay_1,
        assay_2[, colnames(assay_1)]
    )
})

test_that("get_unharmonised_dataset works with one ID", {
    dataset_id = "838ea006-2369-4e2c-b426-b2a744a2b02b"
    unharmonised_meta = get_unharmonised_dataset(dataset_id)

    expect_s3_class(unharmonised_meta, "tbl")
})

test_that("get_unharmonised_metadata() returns the appropriate data", {
    harmonised <- get_metadata() |> dplyr::filter(tissue == "kidney blood vessel")
    unharmonised <- get_unharmonised_metadata(harmonised)
    
    unharmonised |> is.data.frame() |> expect_true()
    expect_setequal(colnames(unharmonised), c("file_id", "unharmonised"))
    
    # The number of cells in both harmonised and unharmonised should be the same
    expect_equal(
        dplyr::collect(harmonised) |> nrow(),
        unharmonised$unharmonised |> purrr::map_int(function(df) dplyr::tally(df) |> dplyr::pull(n)) |> sum()
    )
    
    # The number of datasets in both harmonised and unharmonised should be the same
    expect_equal(
        harmonised |> dplyr::group_by(file_id) |> dplyr::n_groups(),
        nrow(unharmonised)
    )
})

test_that("get_metadata() is cached", {
    table = get_metadata()
    table_2 = get_metadata()
    
    identical(table, table_2) |> expect_true()
})

# Utility scripts for development purposes, that are not exported to users

#' Upload a file to the Nectar object store
#' @param source A character scalar indicating the local path to the file to
#'   upload
#' @param container A character scalar indicating the name of the container to
#'   upload to
#' @param name An optional character scalar indicating the name the file should
#'   have after being uploaded. Defaults to being the basename of the source
#'   file.
#' @param credential_id The OpenStack application credential ID as a character
#'   scalar. This is optional because you can alternatively source a
#'   `-openrc.sh` file instead of providing it here.
#' @param credential_id The OpenStack application credential secret as a
#'   character scalar
#' @return `NULL`, invisibly
#' @keywords internal
upload_swift <- function(
    source,
    container,
    name = basename(source),
    credential_id = NULL,
    credential_secret = NULL
) {
    # Create the basilisk environment
    swift_env <- basilisk::BasiliskEnvironment(
        envname="swift-nectar-upload",
        pkgname=packageName(),
        packages=c(
          "python-swiftclient==4.2.0",
          "python-keystoneclient==5.1.0",
          "python==3.10.9"
        )
    )
    proc <- basilisk::basiliskStart(swift_env)
    
    # Build the CLI args
    if (!is.null(credential_id) && !is.null(credential_secret)){
        auth <- c(
            "--os-auth-type",
            "v3applicationcredential",
            "--os-application-credential-id",
            credential_id,
            "--os-application-credential-secret",
            credential_secret
        )
    }
    else {
        auth <- character()
    }
    args <- c(
        "-m", 
        "swiftclient.shell",
        "--os-auth-url",
        "https://keystone.rc.nectar.org.au:5000/v3/",
        "--os-project-id",
        "06d6e008e3e642da99d806ba3ea629c5",
        auth,
        "upload",
        container,
        source,
        "--object-name",
        name
    )
    
    # Perform the upload
    system2(reticulate::py_exe(), args=args)
    basilisk::basiliskStop(proc)
    
    invisible(NULL)
}

#' Update the metadata database in nectar using a newly created data frame
#' @param metadata The data frame to upload
#' @param version The version for the new metadata as a character scalar, e.g.
#'   "0.2.3"
#' @inheritDotParams upload_swift
#' @examples
#' \dontrun{
#'  metadata = CuratedAtlasQueryR::get_metadata() |>
#'      head(10) |>
#'      dplyr::collect()
#'  update_database(
#'      metadata, 
#'      "0.2.3", 
#'      credential_id = "ABCDEFGHIJK", 
#'      credential_secret = "ABCD1234EFGH-5678IJK"
#'  )
#'  # Prints "metadata.0.2.3.parquet" if successful
#' }
#' @keywords internal
#' @inherit upload_swift return
update_database <- function(metadata, version, ...){
    # These are optional dev packages
    rlang::check_installed(c("arrow", "glue", "basilisk"))
    
    dir <- tempdir()
    parquet_name <- glue::glue("metadata.{version}.parquet")
    parquet_path <- file.path(dir, parquet_name)
    arrow::write_parquet(metadata, sink=parquet_path)
    
    upload_swift(parquet_path, container="metadata", name=parquet_name, ...)
}

#' Update the unharmonised parquet files
#' @param unharmonised_parquet_dir The path to a directory containing parquet
#'   files, one for each dataset, e.g.
#'   /vast/projects/cellxgene_curated/metadata_non_harmonised_parquet_0.2
#' @inheritDotParams upload_swift
#' @inherit upload_swift return
#' @keywords internal
#' @examples
#' \dontrun{
#' update_unharmonised(
#'     "/vast/projects/cellxgene_curated/metadata_non_harmonised_parquet_0.2", 
#'     credential_id = "ABCDEFGHIJK", 
#'     credential_secret = "ABCD1234EFGH-5678IJK"
#' )
#' }
update_unharmonised <- function(unharmonised_parquet_dir, ...){
    # name="/" forces it have no prefix, ie be at the top level in the bucket
    upload_swift(
        unharmonised_parquet_dir, 
        container="unharmonised_metadata",
        name="/", 
        ...
    )
}

#' Converts a series of HDF5Array-serialized SingleCellExperiments to AnnData
#' @param input_directory A character scalar. The path to a directory containing one or more
#'  directories created by [HDF5Array::saveHDF5SummarizedExperiment()].
#' @param output_directory A character scalar. The path to a directory in which to save the
#'  created anndata files.
#' @keywords internal
#' @return A character vector of the newly-created anndata files
#' @examples
#' \donttest{
#' hdf5_to_anndata(
#'     "/vast/projects/cellxgene_curated/splitted_DB2_data_0.2.1",
#'     "/vast/projects/cellxgene_curated/splitted_DB2_anndata_0.2.1"
#' )
#' hdf5_to_anndata(
#'     "/vast/projects/cellxgene_curated/splitted_DB2_data_scaled_0.2.1",
#'     "/vast/projects/cellxgene_curated/splitted_DB2_anndata_scaled_0.2.1"
#' )
#' }
hdf5_to_anndata <- function(input_directory, output_directory){
    dir.create(output_directory, showWarnings = FALSE)
    # This is a quick utility script to convert the SCE files into AnnData format for use in Pythonlist.files("/vast/projects/RCP/human_cell_atlas/splitted_DB2_data", full.names = FALSE) |>  purrr::walk(function(dir){
    basilisk::basiliskRun(fun = function(sce) {
        list.dirs(input_directory)[-1] |>
            purrr::map_chr(function(sce_dir){
                cli::cli_alert_info("Processing {sce_dir}.")
                prefix <- basename(sce_dir)
                out_path <- glue::glue("{prefix}.h5ad") |>
                    file.path(output_directory, name=_)
                
                if (file.exists(out_path)) {
                    cli::cli_alert_info("{out_path} already exists. Skipping")
                }
                else {
                    sce <- HDF5Array::loadHDF5SummarizedExperiment(sce_dir)
                    single_column <- length(colnames(sce)) == 1
                    if (single_column){
                        # Hack, so that single-column SCEs will convert 
                        # correctly
                        cli::cli_alert_info(
                            "{sce_dir} has only 1 column. Duplicating column."
                        )
                        sce <- cbind(sce, sce)
                        single_column <- TRUE
                    }
                    ad <- zellkonverter::SCE2AnnData(sce)
                    if (single_column){
                        # Remove the duplicate column
                        sce$X <- sce$X[1]
                    }
                    # TODO: customize chunking here, when anndata supports it
                    # (see https://github.com/scverse/anndata/issues/961)
                    ad$write_h5ad(out_path)
                }
                out_path
            }, .progress = "Converting files")
    }, env = zellkonverter::zellkonverterAnnDataEnv())
}

# Converts a series of H5-serialized Seurat to AnnData
# @param input_directory A character scalar. The path to a directory containing one or more
#  directories created by [SeuratDisk::SaveH5Seurat()].
# @param output_directory A character scalar. The path to a directory in which to save the
# created anndata files.
# @keywords internal
# @return A character vector of the newly-created anndata files
# @noRd
# @examples
# \donttest{
# h5seurat_to_anndata(
#     "/vast/projects/cellxgene_curated/splitted_DB2_data_0.2.1",
#     "/vast/projects/cellxgene_curated/splitted_DB2_anndata_0.2.1"
# )
# h5seurat_to_anndata(
#     "/vast/projects/cellxgene_curated/splitted_DB2_data_scaled_0.2.1",
#     "/vast/projects/cellxgene_curated/splitted_DB2_anndata_scaled_0.2.1"
# )
# }
# h5seurat_to_anndata <- function(input_directory, output_directory, assays = "RNA"){
#   
#   # Check if package is loaded
#   if(!"SeuratDisk" %in% (.packages()))
#     stop("CuratedCellAtlas says: please manually load the SeuratDisk package first. Execute `library(SeuratDisk)`")
#     
#     
#   dir.create(output_directory, showWarnings = FALSE)
#   # This is a quick utility script to convert the SCE files into AnnData format for use in Pythonlist.files("/vast/projects/RCP/human_cell_atlas/splitted_DB2_data", full.names = FALSE) |>  purrr::walk(function(dir){
#   basilisk::basiliskRun(fun = function(sce) {
#     dir(input_directory, full.names = TRUE) |>
#       purrr::map_chr(function(seurat_file){
#         cli::cli_alert_info("Processing {seurat_file}.")
#         prefix <- basename(seurat_file)
#         out_path <- glue::glue("{prefix}.h5ad") |>
#           file.path(output_directory, name=_)
#         
#         if (file.exists(out_path)) {
#           cli::cli_alert_info("{out_path} already exists. Skipping")
#         }
#         else {
#           sce <- SeuratDisk::LoadH5Seurat(seurat_file, assays = assays) |> 
#             Seurat::as.SingleCellExperiment()
#           
#           single_column <- length(colnames(sce)) == 1
#           if (single_column){
#             # Hack, so that single-column SCEs will convert 
#             # correctly
#             cli::cli_alert_info(
#               "{seurat_file} has only 1 column. Duplicating column."
#             )
#             sce <- cbind(sce, sce)
#             single_column <- TRUE
#           }
#           ad <- zellkonverter::SCE2AnnData(sce)
#           if (single_column){
#             # Remove the duplicate column
#             sce$X <- sce$X[1]
#           }
#           # TODO: customize chunking here, when anndata supports it
#           # (see https://github.com/scverse/anndata/issues/961)
#           ad$write_h5ad(out_path)
#         }
#         out_path
#       }, .progress = "Converting files")
#   }, env = zellkonverter::zellkonverterAnnDataEnv())
# }

#' Makes a "downsampled" metadata file that only contains the minimal data
#' needed to run the vignette.
#' @noRd
#' @param output Character scalar. Path to the output file.
#' @return NULL
#' @keywords internal
downsample_metadata <- function(output = "sample_meta.parquet"){
    metadata <- get_metadata()
    
    # Make a table of rows per dataset
    dataset_sizes <- metadata |>
        dplyr::group_by(.data$file_id_db) |>
        summarise(n = dplyr::n()) |> 
        dplyr::collect()
    
    # Find a minimal set of file_id_dbs we need
    minimal_file_ids <- rlang::exprs(
        # Used by the vignette
        .data$ethnicity == "African" &
            stringr::str_like(.data$assay, "%10x%") &
            .data$tissue == "lung parenchyma" &
            stringr::str_like(.data$cell_type, "%CD4%"),
        .data$cell_type_harmonised == "nk",
        .data$cell_type_harmonised == "cd14 mono",
        .data$tissue == "kidney blood vessel",
        # Used by tests
        .data$file_id_db == "3214d8f8986c1e33a85be5322f2db4a9",
        .data$cell_ == "868417_1"
    ) |>
        purrr::map(function(filter){
            all_ids <- metadata |> 
                dplyr::filter(!!filter) |>
                dplyr::group_by(.data$file_id_db) |>
                dplyr::pull(.data$file_id_db) |> unique()
                
            dataset_sizes |>
                dplyr::filter(.data$file_id_db %in% all_ids) |>
                dplyr::slice_min(n=50, order_by = .data$n) |>
                dplyr::pull(.data$file_id_db)
        }) |>
        purrr::reduce(union)
    
    metadata |>    
        dplyr::filter(.data$file_id_db %in% minimal_file_ids) |>
        dplyr::arrange(.data$file_id_db, .data$sample_) |>
        dplyr::collect() |>
        arrow::write_parquet(output)
    
    NULL
}

# Utility scripts that are used internally by the package at runtime

#' Gets the file size of a number of remote files
#' @param urls A character vector containing URLs
#' @return The file size of each of the files pointed to by the provided URL,
#' in gigabytes, as double vector
#' @importFrom purrr map_dbl
#' @importFrom httr HEAD
#' @keywords internal
url_file_size <- function(urls){
    map_dbl(urls, function(url){
        as.integer(
            HEAD(url)$headers$`content-length` 
        ) / 10^9
    })
}

#' Prints a message indicating the size of a download
#' @inheritParams url_file_size
#' @importFrom cli cli_alert_info
#' @keywords internal
#' @return `NULL`, invisibly
report_file_sizes <- function(urls){
    total_size <- url_file_size(urls) |> 
        sum() |>
        round(digits=2)
    
    "Downloading {length(urls)} file{?s}, totalling {total_size} GB" |>
        cli_alert_info()
    
    invisible(NULL)
}

#' Formats a multi-line string as it it were on one line
#' @param text Any character vector
#' @return The same character vector, with newlines and subsequent whitespace
#'   removed
#' @keywords internal
#' @importFrom stringr str_remove_all
single_line_str <- function(text){
    str_remove_all(text, r"(\n\s*)")
}

#' Returns the default cache directory
#' @return A length one character vector.
#' @importFrom tools R_user_dir
#' @importFrom utils packageName
#' @keywords internal
get_default_cache_dir <- function() {
    packageName() |>
        R_user_dir(
            "cache"
        ) |>
        normalizePath() |>
        suppressWarnings()
}

#' Synchronises a single remote file with a local path
#' @importFrom httr write_disk GET stop_for_status
#' @importFrom cli cli_abort cli_alert_info
#' @return `NULL`, invisibly
#' @keywords internal
sync_remote_file <- function(full_url, output_file, ...) {
    if (!file.exists(output_file)) {
        output_dir <- dirname(output_file)
        dir.create(output_dir,
                   recursive = TRUE,
                   showWarnings = FALSE
        )
        cli_alert_info("Downloading {full_url} to {output_file}")
        
        tryCatch(
            GET(full_url, write_disk(output_file), ...) |> stop_for_status(),
            error = function(e) {
                # Clean up if we had an error
                file.remove(output_file)
                cli_abort("File {full_url} could not be downloaded. {e}")
            }
        )
    }
    invisible(NULL)
}

#' Returns a tibble from a parquet file path
#' 
#' Since dbplyr 2.4.0, raw file paths aren't handled very well
#' See: https://github.com/duckdb/duckdb-r/issues/38
#' Hence the need for this method
#' @importFrom glue glue
#' @importFrom dplyr tbl
#' @importFrom dbplyr sql
#' @return An SQL data frame
#' @keywords internal
read_parquet <- function(conn, path){
    from_clause <- glue("FROM read_parquet('{path}')") |> sql()
    tbl(conn, from_clause)
}

test_that("url_file_size() returns the correct sizes", {
    c(
        "https://ftp.ncbi.nlm.nih.gov/refseq/H_sapiens/annotation/GRCh38_latest/refseq_identifiers/GRCh38_latest_genomic.fna.gz",
        "https://ftp.ncbi.nlm.nih.gov/refseq/H_sapiens/annotation/GRCh37_latest/refseq_identifiers/GRCh37_latest_genomic.fna.gz"
    ) |>
        url_file_size() |>
        expect_equal(c(
            0.973,
            0.944
        ), tolerance = 0.001)
})

# Functions that relate to unharmonised metadata

#' @include utils.R
NULL

#' Base URL for all the unharmonised data
#' @noRd
UNHARMONISED_URL <- single_line_str(
    "https://object-store.rc.nectar.org.au/v1/
    AUTH_06d6e008e3e642da99d806ba3ea629c5/unharmonised_metadata"
)

#' Returns unharmonised metadata for selected datasets.
#'
#' Various metadata fields are *not* common between datasets, so it does not
#' make sense for these to live in the main metadata table. This function is a
#' utility that allows easy fetching of this data if necessary.
#'
#' @param dataset_id A character vector, where each entry is a dataset ID
#'   obtained from the `$file_id` column of the table returned from
#'   [get_metadata()]
#' @param cells An optional character vector of cell IDs. If provided, only
#'   metadata for those cells will be returned.
#' @param conn An optional DuckDB connection object. If provided, it will re-use
#'   the existing connection instead of opening a new one.
#' @param remote_url Optional character vector of length 1. An HTTP URL pointing
#'   to the root URL under which all the unharmonised dataset files are located.
#' @param cache_directory Optional character vector of length 1. A file path on
#'   your local system to a directory (not a file) that will be used to store
#'   the unharmonised metadata files.
#' @importFrom purrr map set_names
#' @importFrom glue glue
#' @importFrom DBI dbConnect
#' @importFrom duckdb duckdb
#' @importFrom dplyr tbl filter
#' @importFrom rlang .data
#' @return A named list, where each name is a dataset file ID, and each value is
#'   a "lazy data frame", ie a `tbl`.
#' @examples
#' \donttest{
#' dataset <- "838ea006-2369-4e2c-b426-b2a744a2b02b"
#' harmonised_meta <- get_metadata() |> 
#'     dplyr::filter(file_id == dataset) |> dplyr::collect()
#' unharmonised_meta <- get_unharmonised_dataset(dataset)
#' unharmonised_tbl <- dplyr::collect(unharmonised_meta[[dataset]])
#' dplyr::left_join(harmonised_meta, unharmonised_tbl, by=c("file_id", "cell_"))
#' }
get_unharmonised_dataset <- function(
    dataset_id,
    cells = NULL,
    conn = duckdb() |> dbConnect(drv = _, read_only = TRUE),
    remote_url = UNHARMONISED_URL,
    cache_directory = get_default_cache_dir()
){
    unharmonised_root <- file.path(
      cache_directory,
      COUNTS_VERSION,
      "unharmonised"
    )
    file_name <- glue::glue("{dataset_id}.parquet")
    local_path <- file.path(unharmonised_root, file_name)
    glue("{remote_url}/{file_name}") |>
        sync_remote_file(
            local_path,
            progress(type = "down", con = stderr())
        )
    
    read_parquet(conn, local_path) |>
        filter(.data$cell_ %in% cells)
}

#' Returns unharmonised metadata for a metadata query
#' @inherit get_unharmonised_dataset description
#' @param metadata A lazy data frame obtained from [get_metadata()], filtered
#'   down to some cells of interest
#' @inheritDotParams get_unharmonised_dataset
#' @return A tibble with two columns:
#'  * `file_id`: the same `file_id` as the main metadata table obtained from
#'    [get_metadata()]
#'  * `unharmonised`: a nested tibble, with one row per cell in the input
#'    `metadata`, containing unharmonised metadata
#' @export
#' @importFrom dplyr group_by summarise filter collect
#' @importFrom rlang .data
#' @importFrom dbplyr remote_con
#' @examples
#' harmonised <- dplyr::filter(get_metadata(), tissue == "kidney blood vessel")
#' unharmonised <- get_unharmonised_metadata(harmonised)
get_unharmonised_metadata <- function(metadata, ...){
    args <- list(...)
    metadata |>
        collect() |>
        group_by(.data$file_id) |>
        summarise(
            unharmonised = list(
              dataset_id=.data$file_id[[1]],
              cells=.data$cell_,
              conn=remote_con(metadata)
            ) |>
                c(args) |> 
                do.call(get_unharmonised_dataset, args=_) |> 
                list()
        )
}

# Functions that relate to the harmonised metadata database

#' @include utils.R
NULL

#' Environment that we use to cache the DuckDB connections
#' @noRd
cache <- rlang::env(
    metadata_table = rlang::env()
)

#' URL pointing to the full metadata file
#' @export
#' @return A character scalar consisting of the URL
#' @examples
#' get_metadata(remote_url = DATABASE_URL)
DATABASE_URL <- single_line_str(
    "https://object-store.rc.nectar.org.au/v1/
    AUTH_06d6e008e3e642da99d806ba3ea629c5/metadata/metadata.0.2.3.parquet"
)

#' URL pointing to the sample metadata file, which is smaller and for test,
#' demonstration, and vignette purposes only
#' @export
#' @return A character scalar consisting of the URL
#' @examples
#' get_metadata(remote_url = SAMPLE_DATABASE_URL)
SAMPLE_DATABASE_URL <- single_line_str(
    "https://object-store.rc.nectar.org.au/v1/
    AUTH_06d6e008e3e642da99d806ba3ea629c5/metadata/
    sample_metadata.0.2.3.parquet"
)

#' Gets the Curated Atlas metadata as a data frame.
#'
#' Downloads a parquet database of the Human Cell Atlas metadata to a local
#' cache, and then opens it as a data frame. It can then be filtered and passed
#' into [get_single_cell_experiment()] to obtain a
#' [`SingleCellExperiment::SingleCellExperiment-class`]
#'
#' @param remote_url Optional character vector of length 1. An HTTP URL pointing
#'   to the location of the parquet database.
#' @param cache_directory Optional character vector of length 1. A file path on
#'   your local system to a directory (not a file) that will be used to store
#'   `metadata.parquet`
#' @param use_cache Optional logical scalar. If `TRUE` (the default), and this
#'   function has been called before with the same parameters, then a cached
#'   reference to the table will be returned. If `FALSE`, a new connection will
#'   be created no matter what.
#' @return A lazy data.frame subclass containing the metadata. You can interact
#'   with this object using most standard dplyr functions. For string matching,
#'   it is recommended that you use `stringr::str_like` to filter character
#'   columns, as `stringr::str_match` will not work.
#' @export
#' @examples
#' library(dplyr)
#' filtered_metadata <- get_metadata() |>
#'     filter(
#'         ethnicity == "African" &
#'             assay %LIKE% "%10x%" &
#'             tissue == "lung parenchyma" &
#'             cell_type %LIKE% "%CD4%"
#'     )
#'
#' @importFrom DBI dbConnect
#' @importFrom duckdb duckdb
#' @importFrom dplyr tbl
#' @importFrom httr progress
#' @importFrom cli cli_alert_info hash_sha256
#' @importFrom glue glue
#'
#' @details
#'
#' The metadata was collected from the Bioconductor package `cellxgenedp`. it's
#' vignette `using_cellxgenedp` provides an overview of the columns in the
#' metadata. The data for which the column `organism_name` included "Homo
#' sapiens" was collected collected from `cellxgenedp`.
#'
#' The columns `dataset_id` and `file_id` link the datasets explorable through
#' `CuratedAtlasQueryR` and `cellxgenedp`to the CELLxGENE portal.
#'
#' Our representation, harmonises the metadata at dataset, sample and cell
#' levels, in a unique coherent database table.
#'
#' Dataset-specific columns (definitions available at cellxgene.cziscience.com)
#' `cell_count`, `collection_id`, `created_at.x`, `created_at.y`,
#' `dataset_deployments`, `dataset_id`, `file_id`, `filename`, `filetype`,
#' `is_primary_data.y`, `is_valid`, `linked_genesets`, `mean_genes_per_cell`,
#' `name`, `published`, `published_at`, `revised_at`, `revision`, `s3_uri`,
#' `schema_version`, `tombstone`, `updated_at.x`, `updated_at.y`,
#' `user_submitted`, `x_normalization`
#'
#' Sample-specific columns (definitions available at cellxgene.cziscience.com)
#'
#' `sample_`, `.sample_name`, `age_days`, `assay`, `assay_ontology_term_id`,
#' `development_stage`, `development_stage_ontology_term_id`, `ethnicity`,
#' `ethnicity_ontology_term_id`, `experiment___`, `organism`,
#' `organism_ontology_term_id`, `sample_placeholder`, `sex`,
#' `sex_ontology_term_id`, `tissue`, `tissue_harmonised`,
#' `tissue_ontology_term_id`, `disease`, `disease_ontology_term_id`,
#' `is_primary_data.x`
#'
#' Cell-specific columns (definitions available at cellxgene.cziscience.com)
#'
#' `cell_`, `cell_type`, `cell_type_ontology_term_idm`, `cell_type_harmonised`,
#' `confidence_class`, `cell_annotation_azimuth_l2`,
#' `cell_annotation_blueprint_singler`
#'
#' Through harmonisation and curation we introduced custom column, not present
#' in the original CELLxGENE metadata
#'
#' - `tissue_harmonised`: a coarser tissue name for better filtering
#' - `age_days`: the number of days corresponding to the age
#' - `cell_type_harmonised`: the consensus call identity (for immune cells)
#'   using the original and three novel annotations using Seurat Azimuth and 
#'   SingleR
#' - `confidence_class`: an ordinal class of how confident
#'   `cell_type_harmonised` is. 1 is complete consensus, 2 is 3 out of four and
#'   so on.
#' - `cell_annotation_azimuth_l2`: Azimuth cell annotation
#' - `cell_annotation_blueprint_singler`: SingleR cell annotation using 
#'   Blueprint reference
#' - `cell_annotation_blueprint_monaco`: SingleR cell annotation using Monaco 
#'   reference
#' - `sample_id_db`: Sample subdivision for internal use
#' - `file_id_db`: File subdivision for internal use
#' - `sample_`: Sample ID
#' - `.sample_name`: How samples were defined
#'
#'
#' **Possible cache path issues**
#'
#' If your default R cache path includes non-standard characters (e.g. dash
#' because of your user or organisation name), the following error can manifest
#'
#' Error in `db_query_fields.DBIConnection()`: ! Can't query fields. Caused by
#' error: ! Parser Error: syntax error at or near "/" LINE 2: FROM
#' /Users/bob/Library/Caches...
#'
#' The solution is to choose a different cache, for example
#'
#' get_metadata(cache_directory = path.expand('~'))
#' 
get_metadata <- function(
    remote_url = DATABASE_URL,
    cache_directory = get_default_cache_dir(),
    use_cache = TRUE
) {
    hash <- c(remote_url, cache_directory) |> paste0(collapse="") |>
        hash_sha256()
    cached_connection <- cache$metadata_table[[hash]]
    if (!is.null(cached_connection) && isTRUE(use_cache)) {
        cached_connection
    }
    else {
        db_path <- file.path(cache_directory, "metadata.0.2.3.parquet")
        
        if (!file.exists(db_path)){
            report_file_sizes(remote_url)
            sync_remote_file(
                remote_url,
                db_path,
                progress(type = "down", con = stderr())
            )
        }
        
        table <- duckdb() |>
            dbConnect(drv = _, read_only = TRUE) |>
            read_parquet(db_path)
        cache$metadata_table[[hash]] <- table
        table
    }
}

---
title: "CuratedAtlasQueryR"
output: "html_vignette"
params:
  demo_metadata: true
vignette: >
  %\VignetteIndexEntry{CuratedAtlasQueryR}
  %\VignetteEncoding{UTF-8}
  %\VignetteEngine{knitr::rmarkdown}
knit: >
    (function(x, ...){
        proj_root <- rprojroot::find_package_root_file() |> normalizePath()
        rmarkdown::render(
            x,
            output_format = "html_document",
            params = list(
                demo_metadata = TRUE
            )
        )
        rmarkdown::render(
            x,
            output_file = "README.md",
            output_format = "github_document",
            output_dir = proj_root,
            knit_root_dir = proj_root,
            params = list(
                demo_metadata = FALSE
            )
        )
    })
---

```{r, include = FALSE}
knitr::opts_chunk$set(
    collapse = TRUE,
    comment = "#>"
)
root_dir <- knitr::opts_knit$get("root.dir")
if (!is.null(root_dir)){
    # This hack fixes the relative image paths. 
    # See https://github.com/rstudio/rmarkdown/issues/2473
    knitr::opts_knit$set(
        output.dir = root_dir
    )
}
proj_root <- rprojroot::find_package_root_file() |> normalizePath() 
# Utility function for figures to force them to have the correct path
find_figure <- function(names){
    rprojroot::find_package_root_file() |>
        file.path("man", "figures", names) 
}
METADATA_URL = if (params$demo_metadata)
        CuratedAtlasQueryR::SAMPLE_DATABASE_URL else
        CuratedAtlasQueryR::DATABASE_URL
```

<!-- badges: start -->
[![Lifecycle:maturing](https://img.shields.io/badge/lifecycle-maturing-blue.svg)](https://www.tidyverse.org/lifecycle/#maturing)
<!-- badges: end -->

`CuratedAtlasQuery` is a query interface that allow the programmatic exploration and retrieval of the harmonised, curated and reannotated CELLxGENE single-cell human cell atlas. Data can be retrieved at cell, sample, or dataset levels based on filtering criteria.

Harmonised data is stored in the ARDC Nectar Research Cloud, and most 
`CuratedAtlasQuery` functions interact with Nectar via web requests, so a 
network connection is required for most functionality.

```{r, echo=FALSE, out.height = c("139px"), out.width = "120x" }
find_figure("logo.png") |> knitr::include_graphics()
```

```{r, echo=FALSE, out.height = c("58px"), out.width = c("155x", "129px", "202px", "219px", "180px")}
c(
  "svcf_logo.jpeg", 
  "czi_logo.png", 
  "bioconductor_logo.jpg",
  "vca_logo.png",
  "nectar_logo.png"
) |> 
    find_figure() |>
    knitr::include_graphics()
```

# Query interface

## Installation

```{r, eval=FALSE}
devtools::install_github("stemangiola/CuratedAtlasQueryR")
```

## Load the package

```{r, message=FALSE, warning=FALSE}
library(CuratedAtlasQueryR)
```

## Load and explore the metadata

### Load the metadata

```{r}
# Note: in real applications you should use the default value of remote_url 
metadata <- get_metadata(remote_url = METADATA_URL)
metadata
```

The `metadata` variable can then be re-used for all subsequent queries.

### Explore the tissue 

```{r}
metadata |>
    dplyr::distinct(tissue, file_id) 
```

## Download single-cell RNA sequencing counts 

### Query raw counts

```{r}
single_cell_counts = 
    metadata |>
    dplyr::filter(
        ethnicity == "African" &
        stringr::str_like(assay, "%10x%") &
        tissue == "lung parenchyma" &
        stringr::str_like(cell_type, "%CD4%")
    ) |>
    get_single_cell_experiment()

single_cell_counts
```

### Query counts scaled per million

This is helpful if just few genes are of interest, as they can be compared across samples.

```{r}
single_cell_counts = 
    metadata |>
    dplyr::filter(
        ethnicity == "African" &
        stringr::str_like(assay, "%10x%") &
        tissue == "lung parenchyma" &
        stringr::str_like(cell_type, "%CD4%")
    ) |>
    get_single_cell_experiment(assays = "cpm")

single_cell_counts
```

### Extract only a subset of genes

```{r}
single_cell_counts = 
    metadata |>
    dplyr::filter(
        ethnicity == "African" &
        stringr::str_like(assay, "%10x%") &
        tissue == "lung parenchyma" &
        stringr::str_like(cell_type, "%CD4%")
    ) |>
    get_single_cell_experiment(assays = "cpm", features = "PUM1")

single_cell_counts
```

### Extract the counts as a Seurat object

This convert the H5 SingleCellExperiment to Seurat so it might take long time and occupy a lot of memory depending on how many cells you are requesting.

```{r}
single_cell_counts_seurat = 
    metadata |>
    dplyr::filter(
        ethnicity == "African" &
        stringr::str_like(assay, "%10x%") &
        tissue == "lung parenchyma" &
        stringr::str_like(cell_type, "%CD4%")
    ) |>
    get_seurat()

single_cell_counts_seurat
```

## Save your `SingleCellExperiment`

The returned `SingleCellExperiment` can be saved with two modalities, as `.rds` or as `HDF5`.

### Saving as RDS (fast saving, slow reading)

Saving as `.rds` has the advantage of being fast, andd the `.rds` file occupies very little disk space as it only stores the links to the files in your cache.

However it has the disadvantage that for big `SingleCellExperiment` objects, which merge a lot of HDF5 from your `get_single_cell_experiment`, the display and manipulation is going to be slow.
In addition, an `.rds` saved in this way is not portable: you will not be able
to share it with other users.

```{r}
single_cell_counts |> saveRDS("single_cell_counts.rds")
```

### Saving as HDF5 (slow saving, fast reading)

Saving as `.hdf5` executes any computation on the `SingleCellExperiment` and writes it to disk as a monolithic `HDF5`.
Once this is done, operations on the `SingleCellExperiment` will be comparatively very fast.
The resulting `.hdf5` file will also be totally portable and sharable.

However this `.hdf5` has the disadvantage of being larger than the corresponding `.rds` as it includes a copy of the count information, and the saving process is going to be slow for large objects.

```{r}
single_cell_counts |> HDF5Array::saveHDF5SummarizedExperiment("single_cell_counts", replace = TRUE)
```

## Visualise gene transcription

We can gather all CD14 monocytes cells and plot the distribution of HLA-A across all tissues

```{r, results='hide'}
suppressPackageStartupMessages({
    library(ggplot2)
})

# Plots with styling
counts <- metadata |>
  # Filter and subset
  dplyr::filter(cell_type_harmonised == "cd14 mono") |>
  dplyr::filter(file_id_db != "c5a05f23f9784a3be3bfa651198a48eb") |> 
  
  # Get counts per million for HCA-A gene
  get_single_cell_experiment(assays = "cpm", features = "HLA-A") |> 
  suppressMessages() |>
  
  # Add feature to table
  tidySingleCellExperiment::join_features("HLA-A", shape = "wide") |> 
    
  # Rank x axis
  tibble::as_tibble()

# Plot by disease
counts |>
  dplyr::with_groups(disease, ~ .x |> dplyr::mutate(median_count = median(`HLA.A`, rm.na=TRUE))) |> 
  
  # Plot
  ggplot(aes(forcats::fct_reorder(disease, median_count,.desc = TRUE), `HLA.A`,color = file_id)) +
  geom_jitter(shape=".") +
    
  # Style
  guides(color="none") +
  scale_y_log10() +
  theme_bw() +
  theme(axis.text.x = element_text(angle = 60, vjust = 1, hjust = 1)) + 
  xlab("Disease") + 
  ggtitle("HLA-A in CD14 monocytes by disease") 
```

```{r echo=FALSE}
find_figure("HLA_A_disease_plot.png") |> knitr::include_graphics()
```

```{r, results='hide'}
# Plot by tissue
counts |> 
  dplyr::with_groups(tissue_harmonised, ~ .x |> dplyr::mutate(median_count = median(`HLA.A`, rm.na=TRUE))) |> 
  
  # Plot
  ggplot(aes(forcats::fct_reorder(tissue_harmonised, median_count,.desc = TRUE), `HLA.A`,color = file_id)) +
  geom_jitter(shape=".") +
    
  # Style
  guides(color="none") +
  scale_y_log10() +
  theme_bw() +
  theme(axis.text.x = element_text(angle = 60, vjust = 1, hjust = 1)) + 
  xlab("Tissue") + 
  ggtitle("HLA-A in CD14 monocytes by tissue") + 
  theme(legend.position = "none")
```
```{r echo=FALSE}
find_figure("HLA_A_tissue_plot.png") |> knitr::include_graphics()
```

```{r}
metadata |> 
    
  # Filter and subset
  dplyr::filter(cell_type_harmonised=="nk") |> 

  # Get counts per million for HCA-A gene 
  get_single_cell_experiment(assays = "cpm", features = "HLA-A") |> 
  suppressMessages() |>

  # Plot
  tidySingleCellExperiment::join_features("HLA-A", shape = "wide") |> 
  ggplot(aes(tissue_harmonised, `HLA.A`, color = file_id)) +
  theme_bw() +
  theme(
      axis.text.x = element_text(angle = 60, vjust = 1, hjust = 1),
      legend.position = "none"
  ) + 
  geom_jitter(shape=".") + 
  xlab("Tissue") + 
  ggtitle("HLA-A in nk cells by tissue")
```

## Obtain Unharmonised Metadata

Various metadata fields are *not* common between datasets, so it does not
make sense for these to live in the main metadata table. However, we can
obtain it using the `get_unharmonised_metadata()` function. This function
returns a data frame with one row per dataset, including the `unharmonised` 
column which contains unharmnised metadata as a nested data frame.

```{r}
harmonised <- metadata |> dplyr::filter(tissue == "kidney blood vessel")
unharmonised <- get_unharmonised_metadata(harmonised)
unharmonised
```

Notice that the columns differ between each dataset's data frame:

```{r}
dplyr::pull(unharmonised) |> head(2)
```

# Cell metadata

Dataset-specific columns (definitions available at cellxgene.cziscience.com)

`cell_count`, `collection_id`, `created_at.x`, `created_at.y`, `dataset_deployments`, `dataset_id`, `file_id`, `filename`, `filetype`, `is_primary_data.y`, `is_valid`, `linked_genesets`, `mean_genes_per_cell`, `name`, `published`, `published_at`, `revised_at`, `revision`, `s3_uri`, `schema_version`, `tombstone`, `updated_at.x`, `updated_at.y`, `user_submitted`, `x_normalization`

Sample-specific columns (definitions available at cellxgene.cziscience.com)

`sample_`, `sample_name`, `age_days`, `assay`, `assay_ontology_term_id`, `development_stage`, `development_stage_ontology_term_id`, `ethnicity`, `ethnicity_ontology_term_id`, `experiment___`, `organism`, `organism_ontology_term_id`, `sample_placeholder`, `sex`, `sex_ontology_term_id`, `tissue`, `tissue_harmonised`, `tissue_ontology_term_id`, `disease`, `disease_ontology_term_id`, `is_primary_data.x`

Cell-specific columns (definitions available at cellxgene.cziscience.com)

`cell_`, `cell_type`, `cell_type_ontology_term_idm`, `cell_type_harmonised`, `confidence_class`, `cell_annotation_azimuth_l2`, `cell_annotation_blueprint_singler` 

Through harmonisation and curation we introduced custom column, not present in the original CELLxGENE metadata

- `tissue_harmonised`: a coarser tissue name for better filtering
- `age_days`: the number of days corresponding to the age
- `cell_type_harmonised`: the consensus call identity (for immune cells) using the original and three novel annotations using Seurat Azimuth and SingleR
- `confidence_class`: an ordinal class of how confident `cell_type_harmonised` is. 1 is complete consensus, 2 is 3 out of four and so on.             
- `cell_annotation_azimuth_l2`: Azimuth cell annotation
- `cell_annotation_blueprint_singler`: SingleR cell annotation using Blueprint reference
- `cell_annotation_blueprint_monaco`: SingleR cell annotation using Monaco reference
- `sample_id_db`: Sample subdivision for internal use
- `file_id_db`: File subdivision for internal use
- `sample_`: Sample ID
- `.sample_name`: How samples were defined

# RNA abundance

The `raw` assay includes RNA abundance in the positive real scale (not transformed with non-linear functions, e.g. log sqrt). Originally CELLxGENE include a mix of scales and transformations specified in the `x_normalization` column.

The `cpm` assay includes counts per million.

# Session Info
```{r}
sessionInfo()
```

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/metadata.R
\name{get_metadata}
\alias{get_metadata}
\title{Gets the Curated Atlas metadata as a data frame.}
\usage{
get_metadata(
  remote_url = DATABASE_URL,
  cache_directory = get_default_cache_dir(),
  use_cache = TRUE
)
}
\arguments{
\item{remote_url}{Optional character vector of length 1. An HTTP URL pointing
to the location of the parquet database.}

\item{cache_directory}{Optional character vector of length 1. A file path on
your local system to a directory (not a file) that will be used to store
\code{metadata.parquet}}

\item{use_cache}{Optional logical scalar. If \code{TRUE} (the default), and this
function has been called before with the same parameters, then a cached
reference to the table will be returned. If \code{FALSE}, a new connection will
be created no matter what.}
}
\value{
A lazy data.frame subclass containing the metadata. You can interact
with this object using most standard dplyr functions. For string matching,
it is recommended that you use \code{stringr::str_like} to filter character
columns, as \code{stringr::str_match} will not work.
}
\description{
Downloads a parquet database of the Human Cell Atlas metadata to a local
cache, and then opens it as a data frame. It can then be filtered and passed
into \code{\link[=get_single_cell_experiment]{get_single_cell_experiment()}} to obtain a
\code{\link[SingleCellExperiment:SingleCellExperiment]{SingleCellExperiment::SingleCellExperiment}}
}
\details{
The metadata was collected from the Bioconductor package \code{cellxgenedp}. it's
vignette \code{using_cellxgenedp} provides an overview of the columns in the
metadata. The data for which the column \code{organism_name} included "Homo
sapiens" was collected collected from \code{cellxgenedp}.

The columns \code{dataset_id} and \code{file_id} link the datasets explorable through
\code{CuratedAtlasQueryR} and \code{cellxgenedp}to the CELLxGENE portal.

Our representation, harmonises the metadata at dataset, sample and cell
levels, in a unique coherent database table.

Dataset-specific columns (definitions available at cellxgene.cziscience.com)
\code{cell_count}, \code{collection_id}, \code{created_at.x}, \code{created_at.y},
\code{dataset_deployments}, \code{dataset_id}, \code{file_id}, \code{filename}, \code{filetype},
\code{is_primary_data.y}, \code{is_valid}, \code{linked_genesets}, \code{mean_genes_per_cell},
\code{name}, \code{published}, \code{published_at}, \code{revised_at}, \code{revision}, \code{s3_uri},
\code{schema_version}, \code{tombstone}, \code{updated_at.x}, \code{updated_at.y},
\code{user_submitted}, \code{x_normalization}

Sample-specific columns (definitions available at cellxgene.cziscience.com)

\code{sample_}, \code{.sample_name}, \code{age_days}, \code{assay}, \code{assay_ontology_term_id},
\code{development_stage}, \code{development_stage_ontology_term_id}, \code{ethnicity},
\code{ethnicity_ontology_term_id}, \code{experiment___}, \code{organism},
\code{organism_ontology_term_id}, \code{sample_placeholder}, \code{sex},
\code{sex_ontology_term_id}, \code{tissue}, \code{tissue_harmonised},
\code{tissue_ontology_term_id}, \code{disease}, \code{disease_ontology_term_id},
\code{is_primary_data.x}

Cell-specific columns (definitions available at cellxgene.cziscience.com)

\code{cell_}, \code{cell_type}, \code{cell_type_ontology_term_idm}, \code{cell_type_harmonised},
\code{confidence_class}, \code{cell_annotation_azimuth_l2},
\code{cell_annotation_blueprint_singler}

Through harmonisation and curation we introduced custom column, not present
in the original CELLxGENE metadata
\itemize{
\item \code{tissue_harmonised}: a coarser tissue name for better filtering
\item \code{age_days}: the number of days corresponding to the age
\item \code{cell_type_harmonised}: the consensus call identity (for immune cells)
using the original and three novel annotations using Seurat Azimuth and
SingleR
\item \code{confidence_class}: an ordinal class of how confident
\code{cell_type_harmonised} is. 1 is complete consensus, 2 is 3 out of four and
so on.
\item \code{cell_annotation_azimuth_l2}: Azimuth cell annotation
\item \code{cell_annotation_blueprint_singler}: SingleR cell annotation using
Blueprint reference
\item \code{cell_annotation_blueprint_monaco}: SingleR cell annotation using Monaco
reference
\item \code{sample_id_db}: Sample subdivision for internal use
\item \code{file_id_db}: File subdivision for internal use
\item \code{sample_}: Sample ID
\item \code{.sample_name}: How samples were defined
}

\strong{Possible cache path issues}

If your default R cache path includes non-standard characters (e.g. dash
because of your user or organisation name), the following error can manifest

Error in \code{db_query_fields.DBIConnection()}: ! Can't query fields. Caused by
error: ! Parser Error: syntax error at or near "/" LINE 2: FROM
/Users/bob/Library/Caches...

The solution is to choose a different cache, for example

get_metadata(cache_directory = path.expand('~'))
}
\examples{
library(dplyr)
filtered_metadata <- get_metadata() |>
    filter(
        ethnicity == "African" &
            assay \%LIKE\% "\%10x\%" &
            tissue == "lung parenchyma" &
            cell_type \%LIKE\% "\%CD4\%"
    )

}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/utils.R
\name{sync_remote_file}
\alias{sync_remote_file}
\title{Synchronises a single remote file with a local path}
\usage{
sync_remote_file(full_url, output_file, ...)
}
\value{
\code{NULL}, invisibly
}
\description{
Synchronises a single remote file with a local path
}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/unharmonised.R
\name{get_unharmonised_dataset}
\alias{get_unharmonised_dataset}
\title{Returns unharmonised metadata for selected datasets.}
\usage{
get_unharmonised_dataset(
  dataset_id,
  cells = NULL,
  conn = dbConnect(drv = duckdb(), read_only = TRUE),
  remote_url = UNHARMONISED_URL,
  cache_directory = get_default_cache_dir()
)
}
\arguments{
\item{dataset_id}{A character vector, where each entry is a dataset ID
obtained from the \verb{$file_id} column of the table returned from
\code{\link[=get_metadata]{get_metadata()}}}

\item{cells}{An optional character vector of cell IDs. If provided, only
metadata for those cells will be returned.}

\item{conn}{An optional DuckDB connection object. If provided, it will re-use
the existing connection instead of opening a new one.}

\item{remote_url}{Optional character vector of length 1. An HTTP URL pointing
to the root URL under which all the unharmonised dataset files are located.}

\item{cache_directory}{Optional character vector of length 1. A file path on
your local system to a directory (not a file) that will be used to store
the unharmonised metadata files.}
}
\value{
A named list, where each name is a dataset file ID, and each value is
a "lazy data frame", ie a \code{tbl}.
}
\description{
Various metadata fields are \emph{not} common between datasets, so it does not
make sense for these to live in the main metadata table. This function is a
utility that allows easy fetching of this data if necessary.
}
\examples{
\donttest{
dataset <- "838ea006-2369-4e2c-b426-b2a744a2b02b"
harmonised_meta <- get_metadata() |> 
    dplyr::filter(file_id == dataset) |> dplyr::collect()
unharmonised_meta <- get_unharmonised_dataset(dataset)
unharmonised_tbl <- dplyr::collect(unharmonised_meta[[dataset]])
dplyr::left_join(harmonised_meta, unharmonised_tbl, by=c("file_id", "cell_"))
}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/unharmonised.R
\name{get_unharmonised_metadata}
\alias{get_unharmonised_metadata}
\title{Returns unharmonised metadata for a metadata query}
\usage{
get_unharmonised_metadata(metadata, ...)
}
\arguments{
\item{metadata}{A lazy data frame obtained from \code{\link[=get_metadata]{get_metadata()}}, filtered
down to some cells of interest}

\item{...}{
  Arguments passed on to \code{\link[=get_unharmonised_dataset]{get_unharmonised_dataset}}
  \describe{
    \item{\code{dataset_id}}{A character vector, where each entry is a dataset ID
obtained from the \verb{$file_id} column of the table returned from
\code{\link[=get_metadata]{get_metadata()}}}
    \item{\code{cells}}{An optional character vector of cell IDs. If provided, only
metadata for those cells will be returned.}
    \item{\code{conn}}{An optional DuckDB connection object. If provided, it will re-use
the existing connection instead of opening a new one.}
    \item{\code{remote_url}}{Optional character vector of length 1. An HTTP URL pointing
to the root URL under which all the unharmonised dataset files are located.}
    \item{\code{cache_directory}}{Optional character vector of length 1. A file path on
your local system to a directory (not a file) that will be used to store
the unharmonised metadata files.}
  }}
}
\value{
A tibble with two columns:
\itemize{
\item \code{file_id}: the same \code{file_id} as the main metadata table obtained from
\code{\link[=get_metadata]{get_metadata()}}
\item \code{unharmonised}: a nested tibble, with one row per cell in the input
\code{metadata}, containing unharmonised metadata
}
}
\description{
Various metadata fields are \emph{not} common between datasets, so it does not
make sense for these to live in the main metadata table. This function is a
utility that allows easy fetching of this data if necessary.
}
\examples{
harmonised <- dplyr::filter(get_metadata(), tissue == "kidney blood vessel")
unharmonised <- get_unharmonised_metadata(harmonised)
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/utils.R
\name{read_parquet}
\alias{read_parquet}
\title{Returns a tibble from a parquet file path}
\usage{
read_parquet(conn, path)
}
\value{
An SQL data frame
}
\description{
Since dbplyr 2.4.0, raw file paths aren't handled very well
See: https://github.com/duckdb/duckdb-r/issues/38
Hence the need for this method
}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/dev.R
\name{update_unharmonised}
\alias{update_unharmonised}
\title{Update the unharmonised parquet files}
\usage{
update_unharmonised(unharmonised_parquet_dir, ...)
}
\arguments{
\item{unharmonised_parquet_dir}{The path to a directory containing parquet
files, one for each dataset, e.g.
/vast/projects/cellxgene_curated/metadata_non_harmonised_parquet_0.2}

\item{...}{
  Arguments passed on to \code{\link[=upload_swift]{upload_swift}}
  \describe{
    \item{\code{source}}{A character scalar indicating the local path to the file to
upload}
    \item{\code{container}}{A character scalar indicating the name of the container to
upload to}
    \item{\code{name}}{An optional character scalar indicating the name the file should
have after being uploaded. Defaults to being the basename of the source
file.}
    \item{\code{credential_id}}{The OpenStack application credential secret as a
character scalar}
  }}
}
\value{
\code{NULL}, invisibly
}
\description{
Update the unharmonised parquet files
}
\examples{
\dontrun{
update_unharmonised(
    "/vast/projects/cellxgene_curated/metadata_non_harmonised_parquet_0.2", 
    credential_id = "ABCDEFGHIJK", 
    credential_secret = "ABCD1234EFGH-5678IJK"
)
}
}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/utils.R
\name{single_line_str}
\alias{single_line_str}
\title{Formats a multi-line string as it it were on one line}
\usage{
single_line_str(text)
}
\arguments{
\item{text}{Any character vector}
}
\value{
The same character vector, with newlines and subsequent whitespace
removed
}
\description{
Formats a multi-line string as it it were on one line
}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/metadata.R
\docType{data}
\name{SAMPLE_DATABASE_URL}
\alias{SAMPLE_DATABASE_URL}
\title{URL pointing to the sample metadata file, which is smaller and for test,
demonstration, and vignette purposes only}
\format{
An object of class \code{character} of length 1.
}
\usage{
SAMPLE_DATABASE_URL
}
\value{
A character scalar consisting of the URL
}
\description{
URL pointing to the sample metadata file, which is smaller and for test,
demonstration, and vignette purposes only
}
\examples{
get_metadata(remote_url = SAMPLE_DATABASE_URL)
}
\keyword{datasets}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/counts.R
\name{get_SingleCellExperiment}
\alias{get_SingleCellExperiment}
\title{Gets a SingleCellExperiment from curated metadata}
\usage{
get_SingleCellExperiment(...)
}
\arguments{
\item{...}{
  Arguments passed on to \code{\link[=get_single_cell_experiment]{get_single_cell_experiment}}
  \describe{
    \item{\code{data}}{A data frame containing, at minimum, a \code{sample_} column, which
corresponds to a single cell sample ID. This can be obtained from the
\code{\link[=get_metadata]{get_metadata()}} function.}
    \item{\code{assays}}{A character vector whose elements must be either "counts"
and/or "cpm", representing the corresponding assay(s) you want to request.
By default only the count assay is downloaded. If you are interested in
comparing a limited amount of genes, the "cpm" assay is more appropriate.}
    \item{\code{repository}}{A character vector of length one. If provided, it should be
an HTTP URL pointing to the location where the single cell data is stored.}
    \item{\code{cache_directory}}{An optional character vector of length one. If
provided, it should indicate a local file path where any remotely accessed
files should be copied.}
    \item{\code{features}}{An optional character vector of features (ie genes) to return
the counts for. By default counts for all features will be returned.}
  }}
}
\value{
A SingleCellExperiment object, with one assay for each value in the
assays argument
}
\description{
Given a data frame of Curated Atlas metadata obtained from \code{\link[=get_metadata]{get_metadata()}},
returns a \code{\link[SingleCellExperiment:SingleCellExperiment]{SingleCellExperiment::SingleCellExperiment}} object
corresponding to the samples in that data frame
}
\examples{
meta <- get_metadata() |> head(2)
sce <- get_single_cell_experiment(meta)

}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/utils.R
\name{url_file_size}
\alias{url_file_size}
\title{Gets the file size of a number of remote files}
\usage{
url_file_size(urls)
}
\arguments{
\item{urls}{A character vector containing URLs}
}
\value{
The file size of each of the files pointed to by the provided URL,
in gigabytes, as double vector
}
\description{
Gets the file size of a number of remote files
}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/dev.R
\name{update_database}
\alias{update_database}
\title{Update the metadata database in nectar using a newly created data frame}
\usage{
update_database(metadata, version, ...)
}
\arguments{
\item{metadata}{The data frame to upload}

\item{version}{The version for the new metadata as a character scalar, e.g.
"0.2.3"}

\item{...}{
  Arguments passed on to \code{\link[=upload_swift]{upload_swift}}
  \describe{
    \item{\code{source}}{A character scalar indicating the local path to the file to
upload}
    \item{\code{container}}{A character scalar indicating the name of the container to
upload to}
    \item{\code{name}}{An optional character scalar indicating the name the file should
have after being uploaded. Defaults to being the basename of the source
file.}
    \item{\code{credential_id}}{The OpenStack application credential secret as a
character scalar}
  }}
}
\value{
\code{NULL}, invisibly
}
\description{
Update the metadata database in nectar using a newly created data frame
}
\examples{
\dontrun{
 metadata = CuratedAtlasQueryR::get_metadata() |>
     head(10) |>
     dplyr::collect()
 update_database(
     metadata, 
     "0.2.3", 
     credential_id = "ABCDEFGHIJK", 
     credential_secret = "ABCD1234EFGH-5678IJK"
 )
 # Prints "metadata.0.2.3.parquet" if successful
}
}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/metadata.R
\docType{data}
\name{DATABASE_URL}
\alias{DATABASE_URL}
\title{URL pointing to the full metadata file}
\format{
An object of class \code{character} of length 1.
}
\usage{
DATABASE_URL
}
\value{
A character scalar consisting of the URL
}
\description{
URL pointing to the full metadata file
}
\examples{
get_metadata(remote_url = DATABASE_URL)
}
\keyword{datasets}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/dev.R
\name{hdf5_to_anndata}
\alias{hdf5_to_anndata}
\title{Converts a series of HDF5Array-serialized SingleCellExperiments to AnnData}
\usage{
hdf5_to_anndata(input_directory, output_directory)
}
\arguments{
\item{input_directory}{A character scalar. The path to a directory containing one or more
directories created by \code{\link[HDF5Array:saveHDF5SummarizedExperiment]{HDF5Array::saveHDF5SummarizedExperiment()}}.}

\item{output_directory}{A character scalar. The path to a directory in which to save the
created anndata files.}
}
\value{
A character vector of the newly-created anndata files
}
\description{
Converts a series of HDF5Array-serialized SingleCellExperiments to AnnData
}
\examples{
\donttest{
hdf5_to_anndata(
    "/vast/projects/cellxgene_curated/splitted_DB2_data_0.2.1",
    "/vast/projects/cellxgene_curated/splitted_DB2_anndata_0.2.1"
)
hdf5_to_anndata(
    "/vast/projects/cellxgene_curated/splitted_DB2_data_scaled_0.2.1",
    "/vast/projects/cellxgene_curated/splitted_DB2_anndata_scaled_0.2.1"
)
}
}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/utils.R
\name{report_file_sizes}
\alias{report_file_sizes}
\title{Prints a message indicating the size of a download}
\usage{
report_file_sizes(urls)
}
\arguments{
\item{urls}{A character vector containing URLs}
}
\value{
\code{NULL}, invisibly
}
\description{
Prints a message indicating the size of a download
}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/utils.R
\name{get_default_cache_dir}
\alias{get_default_cache_dir}
\title{Returns the default cache directory}
\usage{
get_default_cache_dir()
}
\value{
A length one character vector.
}
\description{
Returns the default cache directory
}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/dev.R
\name{upload_swift}
\alias{upload_swift}
\title{Upload a file to the Nectar object store}
\usage{
upload_swift(
  source,
  container,
  name = basename(source),
  credential_id = NULL,
  credential_secret = NULL
)
}
\arguments{
\item{source}{A character scalar indicating the local path to the file to
upload}

\item{container}{A character scalar indicating the name of the container to
upload to}

\item{name}{An optional character scalar indicating the name the file should
have after being uploaded. Defaults to being the basename of the source
file.}

\item{credential_id}{The OpenStack application credential secret as a
character scalar}
}
\value{
\code{NULL}, invisibly
}
\description{
Upload a file to the Nectar object store
}
\keyword{internal}
